# Copyright (C) 2018, 2019, 2020, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage.pytest import *
import os.path
import pytest
from carthage.dependency_injection import *
from carthage.dependency_injection import DependencyProvider
from carthage import base_injector, network, rsync_git_tree, V4Config, vm, Machine, sh
from carthage.network import random_mac_addr
from carthage.config import ConfigLayout
from carthage.vm import VM, vm_image
from carthage.network import NetworkConfig
from carthage.machine import ssh_origin
import carthage.ansible
from carthage.modeling import *
import gc
import posix
import os

resource_dir = os.path.dirname(__file__)

pytestmark = pytest.mark.no_rootless

@pytest.fixture()
def ainjector():
    if posix.geteuid() != 0:
        pytest.skip("Not running as root; volume tests skipped", )
    try:
        sh.virsh
    except BaseException:
        pytest.skip("libvirt not installed")
    injector = base_injector.claim()(AsyncInjector)
    cl = injector.get_instance(InjectionKey(ConfigLayout))
    cl.delete_volumes = True
    nc = NetworkConfig()
    nc.add('eth0', network.external_network_key, None)
    injector.add_provider(nc)
    injector.replace_provider(ssh_origin, DependencyProvider(None))
    yield injector
    gc.collect()


@async_test
async def test_vm_config(loop, ainjector, vm_image):
    vm = await ainjector(VM, name="vm_1", image=vm_image)
    await vm.write_config()


@async_test
async def test_vm_test(request, ainjector, vm_image):
    with TestTiming(300):
        vm = await ainjector(VM, name="vm_2", image=vm_image)
    vm.ssh_rekeyed()
    assert vm.config_layout.delete_volumes
    with TestTiming(400):
        async with vm.machine_running():
            await vm.ssh_online()
            await vm.ssh("apt-get update")
            await vm.ssh("apt-get -y install python3-pytest ansible rsync python3-mako python3-sh python3-lmdb locales-all")
            await ainjector(rsync_git_tree, resource_dir, vm.rsync_path('/carthage'))
            await subtest_controller(request, vm, "/carthage/tests/inner_plugin_test.py",
                                     python_path="/carthage")
            # We also test ansible here because we already have a VM up and running
            await ainjector(
                carthage.ansible.run_playbook,
                ["vm"],
                "/carthage/tests/resources/test_playbook.yml",
                inventory="/carthage/tests/resources/inventory.txt",
                origin=vm)


@async_test
async def test_cloud_init(test_ainjector, vm_image):
    try:
        sh.virsh
    except Exception:
        pytest.skip("libvirt not installed")
    ainjector = test_ainjector
    config = ainjector.injector(ConfigLayout)
    assert config.delete_volumes

    class layout(CarthageLayout):
        name = "test_cloud_init"

        @provides("test_net")
        class test_net(NetworkModel, AsyncInjectable):
            name = "test_net"

            async def async_ready(self):
                await super().async_ready()
                from carthage.network import BridgeNetwork
                net = await self.ainjector(self.access_by, BridgeNetwork)
                sh.ip(
                    "addr",
                    "add", "10.2.0.1/24", "dev", net.bridge_name)

        class net_config(NetworkConfigModel):
            add('eth0', mac=random_mac_addr(), net=InjectionKey("test_net", _ready=True),
                v4_config=V4Config(address="10.2.0.2",
                                   network="10.2.0.0/24"))

        add_provider(machine_implementation_key, dependency_quote(VM))
        add_provider(carthage.vm.vm_image, vm_image)

        class vm_3(MachineModel):
            name = "vm-3"
            ip_address = "10.2.0.2"
            cloud_init = True
    ainjector.add_provider(layout)
    l = await ainjector.get_instance_async(layout)
    ainjector = l.ainjector
    with TestTiming(300):
        m = await ainjector.get_instance_async(InjectionKey(Machine, host="vm-3"))
        m.ssh_rekeyed()
        async with m.machine_running(ssh_online=True):
            pass



# Copyright (C) 2019, 2021, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *
import carthage
from carthage import sh, ssh
config = base_injector(ConfigLayout)
from carthage.dependency_injection.introspection import *

async def async_setup():
    global ssh_key, tmux
#    ssh_agent = await ainjector.get_instance_async(ssh.SshAgent)
#    tmux = sh.tmux.bake('new-window', _env=ssh_agent.agent_environ)
pass


def unbake(cmd):
    "Helper to split out the arguments of a baked sh command"
    return [cmd._path] + list(cmd._partial_baked_args)


async def ssh_to(m: carthage.machine.Machine):
    if in_tmux:
        ssh_key = await ainjector.get_instance_async(carthage.ssh.SshKey)
        sh.tmux("setenv", "-g", "SSH_AUTH_SOCK", ssh_key.agent.auth_sock)
        sh.tmux("setenv", "-u", "SSH_AUTH_SOCK")
        tmux(unbake(m.ssh), '-A')
    else:
        await loop.run_in_executor(None, m.ssh('-A', _fg=True))



# Copyright (C) 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import json
import logging
import os
import re
from pathlib import Path

from .setup_tasks import *
from .dependency_injection import *
from .config import ConfigLayout
from .network import hash_network_links
logger = logging.getLogger('carthage.network')
__all__ = []

port_number_re = re.compile(r'(Ethernet|PortChannel)([0-9]+)')
speed_re = re.compile(r'([0-9]+)[gG]')


def sonic_port_config(model, breakout_config, network_links):
    def config_for(link, breakout_mode):
        nonlocal current_breakout
        nonlocal port_config
        breakout_info = breakout_config[link]
        indexes = breakout_info['index'].split(',')
        lanes = breakout_info['lanes'].split(',')
        if breakout_mode not in breakout_info['breakout_modes']:
            logger.error(f'{model.name}: {breakout_mode} not valid breakout mode for {link}')
            return
        aliases = breakout_info['breakout_modes'][breakout_mode]
        current_breakout[link] = dict(brkout_mode=breakout_mode)
        num_ports = len(aliases)
        num_lanes = int(len(lanes) / num_ports)
        speed = int(speed_re.search(breakout_mode).group(1)) * 1000
        offset = 0
        portbase = int(port_number_re.match(link).group(2))
        # At least on some switches, it's mandatory to have all the
        # breakout ports be the same speed.  Even when not mandatory,
        # that's probably what you want.  So, see if any speed is
        # configured by the NetworkConfig, and if so, use the first
        # such speed as a default.  Later code will allow a link to
        # override.
        for portnum in range(portbase, portbase + num_ports, 1):
            try:
                speed = network_links['Ethernet' + str(portnum)].speed
                break
            except (AttributeError, KeyError):
                continue

        for portnum in range(portbase, portbase + num_ports, 1):
            port_config[f'Ethernet{portnum}'] = dict(
                admin_status='up',
                lanes=",".join(lanes[(offset * num_lanes):((offset + 1) * num_lanes)]),
                index=str(indexes[offset]),
                alias=aliases[offset],
                mtu="9100",
                speed=str(speed),
            )
            offset += 1

    port_config = {}
    current_breakout = {}
    for link in breakout_config:
        breakout_mode = breakout_config[link]['default_brkout_mode']
        if link in network_links:
            try:
                breakout_mode = network_links[link].breakout_mode
            except AttributeError:
                pass
        config_for(link, breakout_mode)
    for link in network_links:
        if not link.startswith('Ethernet'):
            continue
        if network_links[link].local_type:
            continue
        if link not in port_config:
            logger.error(f'{model.name}: {link} configured but not present in physical breakout modes')
            continue
        nl = network_links[link]
        if nl.mtu:
            port_config[link]['mtu'] = str(nl.mtu)
        # Speed defaults to one specified in breakout mode.  Next
        # lowest priority is speed from other side of link, then speed
        # from this side of the link.
        try:
            port_config[link]['speed'] = str(nl.other.speed)
        except AttributeError:
            pass
        try:
            port_config[link]['speed'] = str(nl.speed)
        except AttributeError:
            pass

    return [
        dict(
            op='add',
            path='/PORT',
            value=port_config),
        dict(
            op='add',
            path='/BREAKOUT_CFG',
            value=current_breakout)]


def sonic_portchannel_config(model, network_links):
    def pc(num):
        return f'PortChannel{num}'

    portchannels = {}
    members = {}
    for l in network_links.values():
        if l.local_type:
            continue
        if not getattr(l, 'portchannel_member', None):
            continue
        pc_name = pc(l.portchannel_member)
        portchannels[pc_name] = dict(
            admin_status='up',
            lacp_key='auto',
            min_links='1',
            mtu=str(l.mtu or 9100),
        )
        members[f'{pc_name}|{l.interface}'] = {}
    return [
        dict(op='add',
             path='/PORTCHANNEL',
             value=portchannels),
        dict(
            op='add',
            path='/PORTCHANNEL_MEMBER',
            value=members)]


def sonic_vlan_config(model, network_links):
    from carthage.network.switch import link_vlan_config
    vlans = set()
    vlan_members = {}
    for l in network_links.values():
        if not (
                l.interface.startswith('Ethernet')
                or l.interface.startswith('PortChannel')):
            continue
        interface = l.interface
        try:
            interface = 'PortChannel' + str(l.portchannel_member)
        except AttributeError:
            pass
        vlan_config = link_vlan_config(l)
        if vlan_config is None or not vlan_config.allowed_vlans:
            continue
        for vl in vlan_config.allowed_vlans:
            vlans.add(vl)
            vlan_members[f'Vlan{vl}|{interface}'] = dict(
                tagging_mode=('untagged' if vl == vlan_config.untagged_vlan else 'tagged'))

    return [
        dict(
            op='add',
            path='/VLAN',
            value={f'Vlan{vl}': dict(vlanid=str(vl)) for vl in vlans}),
        dict(
            op='add',
            path='/VLAN_MEMBER',
            value=vlan_members),
    ]


@inject(config_layout=ConfigLayout)
class SonicNetworkModelMixin(SetupTaskMixin, AsyncInjectable):

    '''
A mixin for :class:`AbstractMachineModel` for SONiC network switches.
'''

    @setup_task("Generate SONiC Config patch")
    async def sonic_config(self):
        await self.resolve_networking()
        breakout_json_path = self.stamp_path.joinpath("breakout.json")
        if not breakout_json_path.exists():
            raise SkipSetupTask
        breakout_json = breakout_json_path.read_text()
        breakout_config = json.loads(breakout_json)
        result = sonic_port_config(self, breakout_config, self.network_links)
        result.extend(sonic_portchannel_config(self, self.network_links))
        result.extend(sonic_vlan_config(self, self.network_links))
        result.extend([
            dict(op='add',
                 path='/INTERFACE',
                 value=dict(Ethernet0={}),
                 ),
        ])
        for item in result:
            if not item['op'] == 'add':
                continue
            value = item['value']
            if isinstance(value, dict) and len(value) == 0:
                item['op'] = 'remove'
                del item['value']

        with self.stamp_path.joinpath("carthage-sonic-config.json").open("wt") as f:
            f.write(json.dumps(result,
                               indent=4))

    @sonic_config.hash()
    def sonic_config(self):
        return str(hash_network_links(self.network_links))

    @sonic_config.invalidator()
    def sonic_config(self, last_run):
        breakout_path = self.stamp_path.joinpath("breakout.json")
        try:
            stat = breakout_path.stat()
        except FileNotFoundError:
            return False
        return last_run > stat.st_mtime

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        from carthage.modeling import MachineMixin
        self.injector.add_provider(
            InjectionKey(MachineMixin, name="SonicNetworkInstallMixin"),
            dependency_quote(SonicNetworkInstallMixin))


class SonicNetworkInstallMixin(SetupTaskMixin):

    @setup_task("Get breakout config")
    async def get_sonic_breakout_config(self):
        breakout = await self.ssh("show interface breakout",
                                  _bg=True,
                                  _bg_exc=True,
                                  )
        breakout_path = self.model.stamp_path / "breakout.json"
        with breakout_path.open("wt") as f:
            f.write(str(breakout.stdout, 'UTF-8'))
        await self.model.ainjector(self.model.sonic_config)

    @get_sonic_breakout_config.check_completed()
    def get_sonic_breakout_config(self):
        breakout_path = self.model.stamp_path / "breakout.json"
        try:
            stat = breakout_path.stat()
        except FileNotFoundError:
            return False
        return stat.st_mtime


__all__ += ['SonicNetworkModelMixin']



# Copyright (C) 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import contextlib
from ipaddress import IPv4Address
from pathlib import Path
from .machine import Machine
from .dependency_injection import *
from . import sh
from .utils import memoproperty, when_needed
from .setup_tasks import SetupTaskMixin
from .network import NetworkLink, BridgeNetwork, match_link


class LocalMachineMixin:

    '''A mixin for :class:`machines <Machine>` that represent the locally running system.  If no special behavior is required :class:`LocalMachine` should be used instead.  This mixin is appropriate for cloud machines where it is desirable to retain functionality such as using cloud APIs to examine the configuration, but where  marking the machine as local to prevent shutdown and simplify filesysetm access is desired.

When testing whether a :class:`Machine` is local, test for ``isinstance(machine, LocalMachineMixin)``

    '''

    ip_address = "127.0.0.1"

    @contextlib.asynccontextmanager
    async def filesystem_access(self):
        yield Path("/")

    async def stop_machine(self):
        raise NotImplementedError("Stopping localhost may be more dramatic than desired")

    @property
    def shell(self):
        # We don't actually need to enter a namespace, but this provides similar
        # semantics to what we get with containers
        return sh.nsenter.bake()


class LocalMachine(LocalMachineMixin, Machine, SetupTaskMixin):

    '''A machine representing the node on which carthage is running.
    '''

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.running = True

    async def async_ready(self):
        await self.resolve_networking()
        await self.run_setup_tasks()
        await super().async_ready()

    async def start_machine(self):
        await self.start_dependencies()
        await super().start_machine()
        return

    async def is_machine_running(self):
        self.running = True
        return True

    @memoproperty
    def stamp_path(self):
        return Path(self.config_layout.state_dir + "/localhost")


def process_local_network_config(model):
    '''
    Carthage uses :class:`~BridgeNetwork` to connecgt VMs and containers to networking on the local system.  When a network is contained entirely within one hypervisor, things are easy.  However, if the network configuration interacts with the :class:`~carthage.network.NetworkConfig` of the hypervisor, it's important that networks in NetworkConfigs of VMs on containers match up with networks in the hypervisor's network config.

    One approach to accomplish this is to set the :obj:`~carthage.modeling.NetworkMobdel.bridge_name` property in the :class:`carthage.model.NetworkModel`.

However if Carthage is configuring the local networking on the hypervisor, then the information is already encoded in the :class:`~NetworkLink` for the hypervisor.  This function finds bridge links in a NetworkConfig corresponding to the machine on which Carthage is running and sets the bridge names that will be used so that Carthage uses the right local bridges.

    '''
    def associate_bridge(net, bridge_name):
        if hasattr(net, 'bridge_name'):
            return
        net.bridge_name = bridge_name
        net.ainjector.add_provider(InjectionKey(BridgeNetwork),
                                   when_needed(BridgeNetwork, bridge_name=bridge_name, delete_bridge=False))

    from carthage.network.links import BridgeLink
    import netifaces
    excluded_links = set()
    gateways = netifaces.gateways()
    try: v4_gateway_interface = gateways['default'][netifaces.AF_INET][1]
    except (KeyError, IndexError): v4_gateway_interface = None
    for interface in netifaces.interfaces():
        addresses = netifaces.ifaddresses(interface)
        try: mac = addresses[netifaces.AF_LINK][0]['addr']
        except KeyError: mac = None
        try: address = addresses[netifaces.AF_INET][0]['addr']
        except KeyError: address = None
        link = match_link(model.network_links, interface,
                          mac=mac, address=address,
                          excluded_links=excluded_links)
        if link and address:
            link.merged_v4_config.address = IPv4Address(address)
            if interface == v4_gateway_interface:
                link.merged_v4_config.gateway = IPv4Address(gateways['default'][netifaces.AF_INET][0])
                if mac and link.mac is None: link.mac = mac
            # we could also handle secondary addresses.
            
    for l in model.network_links.values():
        if not isinstance(l, BridgeLink):
            continue
        associate_bridge(l.net, l.interface)


__all__ = ['LocalMachineMixin', 'LocalMachine', 'process_local_network_config']



# Copyright (C)  2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.
__all__ = []

import carthage
from .base import *
__all__ += ['PodmanPod', 'PodmanContainer', 'PodmanImage',
            'PodmanFromScratchImage', 'podman_image_volume_key',
            'image_layer_task', 'ContainerfileImage']

from .modeling import PodmanPodModel, PodmanImageModel, ContainerfileImageModel

__all__ += ['PodmanPodModel', 'PodmanImageModel', 'ContainerfileImageModel']


@carthage.inject(injector=carthage.Injector)
def carthage_plugin(injector):
    injector.add_provider(PodmanNetwork, allow_multiple=True)
    



# Copyright (C) 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import logging
import os
import types
import typing
from pathlib import Path
from .implementation import *
from .decorators import *
from carthage.dependency_injection import *  # type: ignore
from carthage.utils import when_needed, memoproperty
from carthage import ConfigLayout, SetupTaskMixin
import carthage.kvstore
import carthage.network
import carthage.machine
from .utils import *

logger = logging.getLogger(__name__)

__all__ = []


def dependency_quote_class(c: type):
    "When *c* is injected in a model, dependency_quote the result"
    from .implementation import classes_to_quote
    classes_to_quote.add(c)


__all__ += ['dependency_quote_class']


@inject_autokwargs(injector=Injector)
class InjectableModel(Injectable, metaclass=InjectableModelType):

    def __init__(self, *args, _not_transcluded=None, **kwargs):
        self.__class__._already_instantiated = True # no more calling modelmethods
        super().__init__(*args, **kwargs)
        injector = self.injector
        dependency_providers: typing.Mapping[typing.Any, DependencyProvider] = {}
        ignored_keys = set()
        parent_injector = injector.parent_injector
        not_transcluded = set()
        if _not_transcluded: not_transcluded.update(_not_transcluded)
        for k, to_ignore in self.__transclusions__.items():
            if k in not_transcluded: continue
            target = parent_injector.injector_containing(k)
            if target:
                ignored_keys |= to_ignore
                # For each alias that a transcluded item may be known by, set up an injector_xref back to the base transcluded key
                for alias in to_ignore:
                    if alias is k: continue # It's the primary key
                    try: injector.add_provider(alias, injector_xref(None, k))
                    except ExistingProvider: pass
            else: not_transcluded |= to_ignore
        if not_transcluded: self.injector.add_provider(not_transcluded_key, not_transcluded)
        self.ignored_by_transclusion = frozenset(ignored_keys)
        # This is complicated because we want to reuse the same
        # DependencyProvider when registering the same value more than
        # once so that instantiations alias and we don't accidentally
        # get multiple instances of the same type providing related
        # but different keys.
        for k, info in self.__class__.__initial_injections__.items():
            v, options = info
            if k in ignored_keys: continue
            try:
                dp = dependency_providers[v]
            # TypeError: unhashable
            except (KeyError, TypeError):
                dp = DependencyProvider(
                    v, close=options['close'],
                    allow_multiple=options['allow_multiple'])
                try:
                    dependency_providers[v] = dp
                except TypeError:
                    pass
            options = dict(options)
            try:
                self.injector.add_provider(k, dp, replace=True, **options)
            except Exception as e:
                raise RuntimeError(f'Failed registering {v} as provider for {k}') from e

        for c in reversed(self.__class__.__mro__):
            if isinstance(c, ModelingBase) and hasattr(c, '_callbacks'):
                for cb in c._callbacks:
                    cb(self)

    def __init_subclass__(cls, *args, template=False, **kwargs):
        super().__init_subclass__(*args, **kwargs)


__all__ += ['InjectableModel']


class NetworkModel(carthage.Network, InjectableModel, metaclass=ModelingContainer):

    def __init__(self, **kwargs):
        kwargs.update(gather_from_class(self, 'name', 'vlan_id'))
        super().__init__(**kwargs)
        if hasattr(self, 'bridge_name'):
            self.ainjector.add_provider(InjectionKey(carthage.network.BridgeNetwork),
                                        when_needed(carthage.network.BridgeNetwork, bridge_name=self.bridge_name, delete_bridge=False))


__all__ += ['NetworkModel']


class NetworkConfigModelType(InjectableModelType):

    @modelmethod
    def add(cls,  interface, *, mac, **kwargs):
        kwargs['mac'] = mac
        if 'net' not in kwargs:
            raise SyntaxError('net is required')
        if isinstance(kwargs['net'], type):
            # see if we can construct an appropriate injector_access
            net = kwargs['net']
            if issubclass(net, NetworkModel) and hasattr(net, '__provides_dependencies_for__'):
                kwargs['net'] = injector_access(net.__provides_dependencies_for__[0])
            else:
                raise SyntaxError(
                    f'net must be an instance of Network (or InjectionKey) not a {kwargs["net"]}; consider wrapping in injector_access')

        def callback(inst):
            nonlocal kwargs
            keys = kwargs.keys()
            values = key_from_injector_access(*kwargs.values())
            kwargs = {k: v for k, v in zip(keys, values)}
            try:
                inst.add(interface, **kwargs)
            except TypeError as e:
                raise TypeError(f'Error constructing {interface} with arguments {kwargs}') from e
        cls._add_callback(callback)


class NetworkConfigModel(InjectableModel,
                         carthage.network.NetworkConfig,
                         metaclass=NetworkConfigModelType
                         ):
    pass


__all__ += ['NetworkConfigModel']


class ModelGroup(InjectableModel, AsyncInjectable, metaclass=ModelingContainer):

    async def all_models(self, ready=None):
        models = await self.ainjector.filter_instantiate_async(
            carthage.machine.ResolvableModel, ['name'],
            stop_at=self.injector,
            ready=ready)
        return [m[1] for m in models]

    async def resolve_networking(self, force=False):
        if hasattr(self, 'resolve_networking_models') and not force:
            return self.resolve_networking_models

        async def await_futures(pending_futures, event, target, **kwargs):
            if pending_futures:
                await asyncio.gather(*pending_futures)
        if not hasattr(self, 'all_model_tasks'):
            model_tasks = await self.ainjector.filter_instantiate_async(
                ModelTasks, ['name'],
                ready=False)
            self.all_model_tasks = [m[1] for m in model_tasks]
        models = await self.all_models(ready=False)
        with self.injector.event_listener_context(
                InjectionKey(carthage.network.NetworkConfig), "resolved",
                await_futures) as event_futures:
            resolve_model_futures = []
            for m in models:
                resolve_model_futures.append(asyncio.ensure_future(m.resolve_model(force)))
            if resolve_model_futures:
                await asyncio.gather(*resolve_model_futures)
        if event_futures:
            await asyncio.gather(*event_futures)
        self.resolve_networking_models = models
        return models

    def close(self, canceled_futures=None):
        try:
            del self.resolved_networking_models
        except BaseException:
            pass
        super().close(canceled_futures)

    async def generate(self):
        async def cb(m):
            try:
                await m.async_become_ready()
            except Exception:
                logger.exception(f"Error generating for {repr(m)}")
                raise
        models = await self.resolve_networking()
        models += self.all_model_tasks
        futures = []
        for m in models:
            if not isinstance(m, AsyncInjectable):
                continue
            futures.append(asyncio.ensure_future(cb(m)))
        if futures:
            await asyncio.gather(*futures)
        if hasattr(super(), 'generate'):
            await super().generate()

    async def async_ready(self):
        await self.resolve_networking()
        return await super().async_ready()


class Enclave(ModelGroup, metaclass=ModelingContainer):

    domain: str

    @classmethod
    def our_key(self):
        return InjectionKey(Enclave, domain=self.domain)


__all__ += ['ModelGroup', 'Enclave']

machine_implementation_key = InjectionKey(carthage.machine.Machine, role="implementation")

__all__ += ['machine_implementation_key']

dependency_quote_class(carthage.machine.BaseCustomization)


class RoleType(ModelingContainer):

    classes_to_inject = (carthage.machine.BaseCustomization,)

class MachineModelType(RoleType):

    namespace_filters = [_handle_base_customization]

    @staticmethod
    def calc_mixin_key(class_name, ns, bases):
        # if we had the class instantiated, we'd just look at
        # self.name, and make up a name from self.__name__ if
        # self.name is not set.  We have to manually traverse the
        # bases (and thus the mro) since we need to perform this
        # before the class is instantiated.
        name = None
        if 'name' in ns:
            name = ns['name']
        else:
            for b in bases:
                if hasattr(b, 'name'):
                    name = b.name
                    break
        if not name:
            name = class_name.lower()
        if '.' not in name:
            try:
                if ns['domain']:
                    name += '.' + ns['domain']
            except KeyError:
                pass

        return InjectionKey(MachineModelMixin, host=name)

    def __new__(cls, name, bases, ns, mixin_key=None, **kwargs):
        bases = adjust_bases_for_tasks(bases, ns)
        template = kwargs.get('template', False)
        if not template:
            if mixin_key is None:
                mixin_key = cls.calc_mixin_key(name, ns, bases)
            try:
                if mixin_key:
                    mixin = ns.get_injected(mixin_key)
                    bases += (mixin,)
            except KeyError:
                pass
        domain = ns.get('domain', None)
        self = super().__new__(cls, name, bases, ns, **kwargs)
        if not template:
            if not hasattr(self, 'name'):
                self.name = self.__name__.lower()
            if domain and not '.' in self.name:
                self.name += '.' + domain
        return self

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not kwargs.get('template', False):
            machine_key = InjectionKey(carthage.machine.Machine, host=self.name, _globally_unique=True)
            self.__transclusions__[machine_key] = frozenset({machine_key})
            self.__transclusion_key__ = self.our_key()
            self.__initial_injections__[machine_key] = (
                self.machine, dict(
                    close=True, allow_multiple=False,
                ))
            self.__container_propagations__[machine_key] = \
                self.__initial_injections__[machine_key]
            globally_unique_key(InjectionKey(
                carthage.machine.ResolvableModel,
                name=self.name))(self)


class MachineModelMixin:
    pass


@inject(
    _not_transcluded=not_transcluded_key)
@inject_autokwargs(
    config_layout=ConfigLayout,
                   )
class MachineModel(InjectableModel, carthage.machine.AbstractMachineModel, metaclass=MachineModelType, template=True):

    '''

    Represents the aspects of a :class:`~carthage.machine.Machine` that are independent of the implementation of that machine.  Typically includes things like:

    * Network configuration (:class:`~carthage.network.NetworkConfig`

    * Configuration for devops tools like Ansible

    * Selecting the target implementation (whether the machine will be a VM, container, or hosted on real hardware)

    Applications that want to reason about the abstract environment typically only need to instantiate models.  Applications that want to build VMs or effect changes to real hardware instantiate the machine implementations.  This class is the modeling extensions to :class:`carthage.machine.AbstractMachineModel`.

    If a *MachineModel* contains reference to :func:`setup_tasks <carthage.setup_task.setup_task>`, then it will automatically gain :class:`~carthage.setup_task.SetupTaskMixin` as a base class.  Similarly, if :func:`~carthage.modeling.decorators.model_mixin_for` is used to decorate a class in the same :class:`CarthageLayout` encountered before the *MachineModel*, then that class will be added as an implicit base class of the model.

    Class Parameters passed in as keywords on the class statement:

    :param template: True if this class represents a base class or mixin rather than a actual model of a specific machine.

    :param mixin_key: The :class:`InjectionKey` to use to search for mixins.


    Any :class:`carthage.network.NetworkConfig` present in the *MachineModel* will be used as the network configuration.

Every :class:`carthage.machine.BaseCustomization` (including MachineCustomization, FilesystemCustomization and ContainerCustomization) will be integrated into the resulting machine:

    * If the customization is called *cust*, then a method will be added to the machine *cust_task* which is a :func:`carthage.machine.customization_task` to run the customization.

    * On the model, *cust*  will become a  method that will produce an instance of the customization applied to the machine.

    For example::

        class server(MachineModel):

            class install_software(MachineCustomization):

                webserver_role = ansible_role_task('webserver')

                database_role = ansible_role_task('database')


    Then *server.machine* will have a method *install_software_task* which will run both ansible roles assuming they have not already been run.  *model.install_software* will produce an instance of the customization applied to *model.machine*.  *model.install_software.database_role()* is a method that will force the database_role to run even if it appears up to date.

    '''

    @classmethod
    def our_key(cls):
        "Returns the globally unique InjectionKey by which this model is known."
        return InjectionKey(MachineModel, host=cls.name, _globally_unique=True)

    def __repr__(self):
        return f'<{self.__class__.__name__} model name: {self.name}>'

    network_config = injector_access(InjectionKey(carthage.network.NetworkConfig))

    #: A set of ansible groups to add a model to; see :func:`carthage.modeling.ansible.enable_modeling_ansible`.
    ansible_groups: typing.Sequence[str]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.network_links = {}
        self.injector.add_provider(InjectionKey(MachineModel), dependency_quote(self))
        self.injector.add_provider(InjectionKey(carthage.machine.AbstractMachineModel), dependency_quote(self))
        machine_key = InjectionKey(carthage.machine.Machine, host=self.name)
        target = self.injector.injector_containing(machine_key)
        if target is self.injector:
            try:
                self.injector.add_provider(InjectionKey(carthage.machine.Machine), MachineImplementation)
            except ExistingProvider:
                raise SyntaxError('carthage.machine.Machine already registered; typically this means a missing dependency_quote when setting machine_implementation_key')
        else:
            self.injector.add_provider(InjectionKey(carthage.machine.Machine), injector_access(machine_key))

    machine = injector_access(InjectionKey(carthage.machine.Machine))

    #: Sequence of classes to be mixed into the resulting machine implementation
    machine_mixins = tuple()

    @memoproperty
    def machine_type(self):
        try:
            implementation = self.injector.get_instance(machine_implementation_key)
        except AsyncRequired:
            raise AsyncRequired(
                'A provider registered for machine_implementation_key has asynchronous dependencies; did you forget a dependency_quote()')
        bases = [implementation] + list(map(lambda x: x[1], self.injector.filter_instantiate(MachineMixin, ['name'])))
        bases += self.machine_mixins
        for b in bases:
            assert isinstance(
                b, type) or hasattr(
                b, '__mro_entries__'), f'{b} is not a type; did you forget a dependency_quote'
        try: res = types.new_class(implementation.__qualname__, tuple(bases))
        except TypeError as e:
            raise TypeError(f'Unable to create machine_type for {self.name} with bases {bases}: {str(e)}') from None
        inject()(res)  # Pick up any injections from extra bases
        for k, customization in self.injector.filter_instantiate(carthage.machine.BaseCustomization, [
                                                                 'description'], stop_at=self.injector):
            name = customization.__name__
            task = carthage.machine.customization_task(customization)
            setattr(res, f'{name}_task', task)
            task.__set_name__(res, name)

        res.model = self
        return res

    @memoproperty
    def stamp_path(self):
        path = self.config_layout.output_dir + f"/hosts/{self.name}"
        os.makedirs(path, exist_ok=True)
        return Path(path)

    async def resolve_networking(self, *args, **kwargs):
        '''
        See :meth:`~carthage.machine.AbstractMachineModel.resolve_networking` for documentation.

        In adition to the standard behavior, if  :meth:`machine_type` is an instance of :class:`~carthage.local.LocalMachineMixin`,
then call :func:`carthage.local.process_local_network_config` to learn about local bridges.
        '''
        res = await super().resolve_networking(*args, **kwargs)
        from carthage.local import LocalMachineMixin, process_local_network_config, LocalMachine
        try:
            if issubclass(self.machine_type, LocalMachineMixin):
                process_local_network_config(self)
        except KeyError:
            pass  # no machine_implementation_key
        return res


@inject(injector=Injector,
        model=MachineModel,
        )
class MachineImplementation(AsyncInjectable):

    # Another class that is only a type because of how the injection
    # machineary works.

    def __new__(cls, injector, model):
        res = model.machine_type
        try:
            return cls.prep(injector(res, name=model.name), model)
        except AsyncRequired:
            self = super().__new__(cls)
            self.name = model.name
            self.model = model
            self.injector = injector
            self.res = res
            return self

    @staticmethod
    def prep(implementation: carthage.machine.Machine, model: MachineModel):
        implementation.model = model
        try:
            implementation.short_name = model.short_name
        except AttributeError:
            pass
        return implementation

    async def async_resolve(self):
        return self.prep(await self.ainjector(self.res, name=self.name), self.model)


__all__ += ['MachineModel', 'MachineModelMixin']

class ImageRole(InjectableModel, metaclass=RoleType): pass

__all__ += ['ImageRole']


class CarthageLayout(ModelGroup):

    '''
    *CarthageLayout* is typically the top level class in a set of Carthage models.  It is a :class:`ModelGroup` that represents a complete collection of objects modeled by Carthage.  The primary purpose of this class is to signify the top of a collection of objects; the layout is generally the place to start when examining a collection of models.

    However, a layout differs from a *ModelGroup* in two ways:

    #. :ref:`carthage-runner <carthage_runner>` looks for a :class:`CarthageLayout` to instantiate after loading plugins.  If the console is used, the layout is made available in the *layout* local variable of the console.  If a command is run, the command is run in the context of the layout.

    #. Layouts that set the `carthage.kvstore.persistent_seed_path` in the context of their :class:`Injector` will have persistent assignments of things like IP addresses and MAC addresses loaded from the seed path when instantiated.

    '''


    @classmethod
    def default_class_injection_key(cls):
        if cls.layout_name:
            return InjectionKey(CarthageLayout, layout_name=cls.layout_name)
        else:
            return InjectionKey(CarthageLayout)

    layout_name = None

    def __init__(self,  **kwargs):
        super().__init__(**kwargs)
        self.injector.add_provider(InjectionKey(CarthageLayout), dependency_quote(self))
        persistent_seed_path = self.injector.get_instance(InjectionKey(carthage.kvstore.persistent_seed_path, _optional=True))
        if persistent_seed_path:
            seed_path = Path(persistent_seed_path)
            if seed_path.exists():
                kvstore = self.injector.get_instance(carthage.kvstore.KvStore)
                if not kvstore.persistent_seed_path:
                    kvstore.load(str(seed_path))


__all__ += ['CarthageLayout']


@inject(ainjector=AsyncInjector)
async def instantiate_layout(layout_name=None, *, ainjector, optional=False):
    if layout_name:
        layout = await ainjector.get_instance_async(InjectionKey(CarthageLayout, layout_name=layout_name, _optional=optional))
    else:
        layout = await ainjector.get_instance_async(InjectionKey(CarthageLayout, _optional=optional))
    return layout

__all__ += ['instantiate_layout']


@inject(injector=Injector)
def model_bases(host: str, *bases,
                injector):
    '''

    One common modeling pattern is to automatically generate  :class:`MachineModel`s for a number of systems from some sort of inventory database.  However, it is often desirable to add a override mechanism so that customizations can be added for a specific model.  This function adds a modeling mixin if one is registered in the injector.  In the automated code it is used like::

        @dynamic_name(inventory.name)
        class model( *injector(model_bases,   inventory.fqdn, MachineModel)):
            name = inventory.fqdn

    And then to add an override somewhere in an injector that :func:`transcludes <transclude_injector>` the model::

        @model_mixin_for(host = "foo.com")
        class FooMixin(MachineModel):
            #This will be a base of the foo.com model

    Compare and contrast with using :func:`transclude_overrides` which will replace the overridden model rather than augmenting it with a mixin.

    '''
    new_bases = list(bases)
    try:
        new_base = injector.get_instance(
            InjectionKey(MachineModelMixin, host=host, ))
        if issubclass(new_base, bases):
            new_bases.insert(0, new_base)
        else:
            new_bases.append(new_base)
    except KeyError:
        pass
    return tuple(new_bases)


__all__ += ['model_bases']


@inject(config_layout=ConfigLayout)
class ModelTasks(InjectableModel, SetupTaskMixin, AsyncInjectable, metaclass=ModelingContainer):

    '''
    A grouping of tasks that will be run at generate time in a :class:`CarthageLayout`.  As part of :meth:`ModelGroup.generate`, the layout searches for any :class:`ModelTasks` provided by its injector and instantiates them.  This causes any setup_tasks to be run.

    All :class:`ModelTasks` have a name, which forms part of their key.  If there needs to be an ordering between tasks, the tasks can inject a dependency on other ModelTasks.

    Example usage::

        class layout(CarthageLayout):

            class mt1(ModelTasks):

                @setup_task("some task")
                def some_task(self): # do stuff

    The *async_ready* method will only be called during generate.  However :class:`ModelTasks` will be instantiated whenever at least :meth:`resolve_networking` is called.

    '''

    @classmethod
    def our_key(cls):
        name = getattr(cls, 'name', cls.__name__)
        return InjectionKey(ModelTasks, name=name)

    @memoproperty
    def stamp_path(self):
        name = getattr(self.__class__, 'name', self.__class__.__name__)
        return Path(self.config_layout.output_dir) / name


__all__ += ['ModelTasks']



# Copyright (C) 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from .base import MachineModel, InjectableModel
from carthage import Machine, when_needed, ConfigLayout
from ..ansible import AnsibleGroupPlugin, AnsibleHostPlugin, AnsibleInventory
from ..dependency_injection import *


class ModelingGroupPlugin(AnsibleGroupPlugin):

    name = 'modeling'

    async def groups_for(self, m):
        if hasattr(m, 'model'):
            return getattr(m.model, 'ansible_groups', [])
        return []

    async def group_info(self):
        return {}


class ModelingHostPlugin(AnsibleHostPlugin):

    name = 'modeling'

    async def host_vars(self, m: Machine):
        try:
            model = m.model
        except BaseException:
            return {}
        return getattr(model, 'ansible_vars', {})


def enable_modeling_ansible(injector: Injector):
    injector.add_provider(InjectionKey(AnsibleGroupPlugin, name='modeling'), ModelingGroupPlugin)
    injector.add_provider(InjectionKey(AnsibleHostPlugin, name='modeling'), ModelingHostPlugin)



class AnsibleModelMixin(InjectableModel, AsyncInjectable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.config_layout = self.injector(ConfigLayout)
        self.injector.add_provider(
            InjectionKey(AnsibleInventory),
            when_needed(AnsibleInventory,
                        destination=self.config_layout.output_dir + "/inventory.yml"))
        enable_modeling_ansible(self.injector)

    async def generate(self):
        await self.ainjector.get_instance_async(AnsibleInventory)


__all__ = ['enable_modeling_ansible', 'AnsibleModelMixin']



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from .base import *
from .decorators import *
from carthage.network import NetworkConfig
from carthage import InjectionKey


class RouterConfig(NetworkConfigModel):

    internet = injector_access("internet")
    site_network = injector_access("site-network")
    add("eth0", net=internet, mac=None)
    add("eth1", net=site_network, mac=None)


class Layout(ModelGroup):

    class net_config(NetworkConfigModel):
        site_network = injector_access("site-network")
        add("eth0", net=site_network, mac=None)

    @provides(InjectionKey("internet"))
    class Internet(NetworkModel):

        #        bridge_name = "brint"
        pass

    class Red(Enclave):
        domain = "evil.com"

        @provides("site-network")
        class RedNet(NetworkModel):
            pass

        class router(MachineModel):
            add_provider(InjectionKey(NetworkConfig), RouterConfig)

        class samba(MachineModel):
            ansible_groups = ['samba']

        for u in ('george', 'sue', 'pat'):
            @dynamic_name(f'{u}_desktop')
            class desktop(MachineModel):
                name = f'{u}-desktop'
        del u



# Copyright (C) 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import collections.abc
import contextlib
import os
import random
import yaml
from pathlib import Path
from ..dependency_injection import *
from ..config import ConfigLayout
import carthage.kvstore


def random_mac_addr():
    mac = [random.randint(0, 255) for x in range(6)]
    mac[0] &= 0xfc  # Make it locally administered
    macstr = [format(m, "02x") for m in mac]
    return ":".join(macstr)


__all__ = ['random_mac_addr']


@inject_autokwargs(config_layout=ConfigLayout,
                   injector=Injector,
                   kvstore=carthage.kvstore.KvStore)
class MacStore(Injectable):

    # If dealing with potential MAC conflicts is important, then
    # rather than using KvStore directly, rewrite to use
    # HintedAssignments.

    def __init__(self, **kwargs):
        from .base import NetworkConfig
        super().__init__(**kwargs)
        state_dir = Path(self.config_layout.state_dir)
        self.path = state_dir / "macs.yml"
        self.domain = self.kvstore.domain('mac', True)
        self.load()
        

    def load(self):
        def recurse(current, base_key):
            for k, v in current.items():
                if isinstance(v, dict):
                    recurse(v, base_key + (k,))
                else:
                    self[base_key + (k,)] = v

        if self.path.exists():
            yaml_dict = yaml.safe_load(self.path.read_text())
            assert isinstance(yaml_dict, dict)
            recurse(yaml_dict, tuple())


    def __contains__(self, k):
        k = self.handle_tuple_key(k)
        res = self.domain.get(k)
        return True if res else False
    
    def __getitem__(self, k):
        k = self.handle_tuple_key(k)
        res = self.domain.get(k)
        if res: return res
        res = random_mac_addr()
        self[k] = res
        return res

    def __setitem__(self, k, v):
        k = self.handle_tuple_key(k)
        self.domain.put(k, v, overwrite=True)
        return v

    def handle_tuple_key(self, k):
        if isinstance(k, str): return k
        if not isinstance(k, collections.abc.Sequence): return k
        k_new = tuple(map( lambda component: component.replace('|','||'), k))
        return "|".join(k)
    
__all__ += ['MacStore']

from ..machine import AbstractMachineModel
from ..local import LocalMachineMixin


@inject(model=AbstractMachineModel,
        store=MacStore)
def persistent_random_mac_always(interface, model, store):
    '''Return a persistant random mac address, even for LocalMachine'''
    return store[(model.name, interface)]

__all__ += ['persistent_random_mac_always']

@inject(model=AbstractMachineModel,
        store=MacStore)
def persistent_random_mac(interface, model, store):
    '''Return a persistent random MAC address if one is already stored, or if the machine type is not a local machine.  In other words, unless the database is pre-seeded, return None for LocalMachine.'''
    if (model.name, interface) in store: return store[(model.name, interface)]
    try:
        machine_type = model.machine_type
        if issubclass(machine_type, LocalMachineMixin): return None
    except AttributeError: pass
    return store[(model.name, interface)]


__all__ += ['persistent_random_mac']


def find_mac_first_member(link):
    '''Return the MAC Address of the first member of this link. Typically this function is not called directly but instead is called when a bridge, bond or VLAN link's *mac* property is set to ``inherit``.
'''
    if link.mac != 'inherit':
        raise ValueError("Intended to be used on alink whose MAC is inherit")
    orig_link = link
    while link:
        if not link.member_links:
            raise ValueError(f"Could not find MAC for {orig_link}")
        link = link.member_links[0]
        if link.mac != 'inherit':
            return link.mac


__all__ += ['find_mac_first_member']



# Copyright (C) 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import enum
import functools
import inspect
import threading
import typing
import warnings
from carthage.dependency_injection import *
from carthage.dependency_injection.base import default_injection_key  # not part of public api
from carthage.dependency_injection.base import InjectorXrefMarker
from .utils import *
from carthage.network import NetworkConfig
import carthage.machine
import carthage.console
# There is a circular import of decorators at the end.

thread_local = threading.local()

__all__ = []


class NSFlags(enum.Flag):
    close = 1
    allow_multiple = 2
    inject_by_name = 4
    inject_by_class = 8
    instantiate_on_access = 16
    propagate_key = 32
    dependency_quote = 64


classes_to_quote = set()


class NsEntry:

    __slots__ = ['extra_keys', 'value',
                 'flags', 'new_name', 'transclusion_key']
    flags: NSFlags
    extra_keys: list
    new_name: typing.Optional[str]
    transclusion_key: typing.Optional[InjectionKey]

    def __init__(self, value):
        self.value = value
        self.extra_keys = []
        self.flags = NSFlags.close | NSFlags.instantiate_on_access | NSFlags.inject_by_name | NSFlags.inject_by_class
        if isinstance(value, type) and (classes_to_quote & set(value.__mro__)):
            self.flags |= NSFlags.dependency_quote
        self.new_name = None
        self.transclusion_key = None

    @property
    def injection_options(self):
        f = self.flags
        d = dict(
            allow_multiple=bool(f & NSFlags.allow_multiple),
            close=bool(f & NSFlags.close),
        )
        return d

    def __repr__(self):
        return f'<NsEntry: flags = {self.flags}, keys: {self.extra_keys}, value: {self.value}>'

    def instantiate_value(self, name):
        if (self.flags & NSFlags.instantiate_on_access) and \
           (self.transclusion_key or isinstance(self.value, type)):
            if self.transclusion_key:
                key = self.transclusion_key
            elif self.extra_keys:
                key = self.extra_keys[0]
            elif self.flags & NSFlags.inject_by_name:
                key = InjectionKey(name)
            else:
                return self.value
            return decorators.injector_access(key, self.value)
        return self.value


class ModelingNamespace(dict):

    '''A dict used as the class namespace for modeling objects.  Allows overrides for:

    * filters to change the value or name that an item is injected under

    * Handling managing InjectionKeys

    '''

    to_inject: typing.Dict[InjectionKey, typing.Tuple[typing.Any, dict]]
    #: A mapping from keys that can be transcluded in this injector to a set of keys in self.to_inject that should not be provided by the injector in an :class:`InjectableModelType` if the mapped key is transcluded.
    transclusions: typing.Mapping[InjectionKey, typing.Set[InjectionKey]]
    initial: typing.Dict[str, typing.Any]

    def __init__(self, cls: type,
                 filters: typing.List[typing.Callable],
                 initial: typing.Mapping,
                 classes_to_inject: typing.Sequence[type]):
        if not hasattr(thread_local, 'current_context'):
            thread_local.current_context = None
        self.cls = cls
        self.filters = list(reversed(filters))
        self.classes_to_inject = frozenset(classes_to_inject)
        self.transclusions = dict()
        super().__init__()
        self.initial = {}
        for k, v in initial.items():
            if k.startswith('_'):
                self.setitem(k,v)
            elif isinstance(v, modelmethod):
                v = functools.partial(v,  cls, self)
            self.initial[k] = v
        self.parent_context = thread_local.current_context
        self.context_imported = False
        self.to_inject = dict()


    def keys_for(self, name, state):
        # returns key, (value, options)
        def val(k):
            value = state.value
            if state.flags & NSFlags.dependency_quote:
                return dependency_quote(value)
            else:
                return value

        options = state.injection_options
        state.extra_keys.sort(key=lambda k:k.globally_unique, reverse=True)
        if state.flags & NSFlags.inject_by_name:
            name_key = InjectionKey(name)
            # If name_key is in extra_keys, use that, so we avoid
            # differing in _globally_unique or _ready
            if name_key not in state.extra_keys:
                yield InjectionKey(name), (val(InjectionKey(name)), options)
        if isinstance(state.value, type) and (state.flags & NSFlags.inject_by_class):
            for b in state.value.__mro__:
                if b in self.classes_to_inject:
                    try:
                        class_key = state.value.default_class_injection_key()
                    except AttributeError:
                        class_key = InjectionKey(b)
                    class_key = InjectionKey(b, **class_key.constraints)
                    yield class_key, (val(class_key), options)
        for k in state.extra_keys:
            yield k, (val(k), options)

    def __getitem__(self, k):
        # Normally when we set an item we drop it from initial
        # but there are some cases where we want a value for getitem different than the value that ends up in the eventual class
        # the obvious case is injector_access from the instantiate flag.
        # so initial actually takes precidence over our actual values.
        try:
            return self.initial[k]
        except KeyError:
            return super().__getitem__(k)

    def __contains__(self, k):
        return super().__contains__(k) or k in self.initial

    def __delitem__(self, k):
        del self[k]
        try:
            del self.to_inject[k]
        except BaseException:
            pass

    def __setitem__(self, k, v):
        if thread_local.current_context is not self:
            self.update_context()
        state = NsEntry(v)
        handled = False
        for f in self.filters:
            if f(self.cls, self, k, state):
                # The filter has handled things
                handled = True
            if state.new_name:
                k = state.new_name
                state.new_name = None
        else:
            if not handled:
                iv = state.instantiate_value(k)
                super().__setitem__(k, iv)
                if iv is not state.value:
                    self.initial[k] = state.value
                else:
                    try:
                        del self.initial[k]
                    except KeyError:
                        pass
        if handled:
            try:
                del self.initial[k]
            except KeyError:
                pass
        if k.startswith('_'):
            if k == "__qualname__":
                self.import_context()
            return state.value
        transclusion_keys = None
        if state.transclusion_key: transclusion_keys = {state.transclusion_key}
        for k, info in self.keys_for(name=k, state=state):
            self.to_inject[k] = info
            if transclusion_keys: transclusion_keys.add(k)
        if transclusion_keys:
            self.transclusions[state.transclusion_key] = frozenset(transclusion_keys)
        return state.value

    def __delitem__(self, k):
        res = super().__delitem__(k)
        try:
            del self.to_inject[InjectionKey(k)]
        except Exception:
            pass
        return res

    def setitem(self, k, v, explicit: bool = True):
        '''An override for use in filters to avoid recursive filtering'''
        res = super().__setitem__(k, v)
        if explicit:
            try:
                del self.initial[k]
            except KeyError:
                pass
        return res

    def keys(self):
        yield from super().keys()
        for k in self.initial.keys():
            if not super().__contains__(k):
                yield k

    def get(self, k, default):
        try:
            return self.initial[k]
        except KeyError:
            return super().get(k, default)

    def import_context(self):
        # We only want to import the context if our qualname is an
        # extension of their qualname, so we need to wait until the
        # first time our qualname is set.  If our qualname is later
        # adjusted we do not want to reimport.
        if (self.parent_context is None) or self.context_imported:
            return
        our_qualname = self['__qualname__']
        parent = self.parent_context
        parent_qualname = parent['__qualname__']
        our_qualname_count = len(our_qualname.split("."))
        parent_qualname_count = len(parent_qualname.split("."))
        self.context_imported = True
        if not (our_qualname.startswith(parent_qualname) and our_qualname_count == parent_qualname_count + 1):
            self.parent_context = None  # Also break injection key lookups at this point
            return
        for k in parent.keys():
            if k in self:
                continue
            if k.startswith('_'):
                continue
            self.initial[k] = parent[k]

    def update_context(self):
        thread_local.current_context = self

    def get_injected(self, key):
        parent = self
        while parent is not None:
            if key in parent.to_inject:
                res = parent.to_inject[key][0]
                if isinstance(res, dependency_quote):
                    res = res.value
                return res
            parent = parent.parent_context
        raise KeyError

def check_already_provided(metaclass_proxy, v):
    if not hasattr(metaclass_proxy, '__namespace__'):
        return None
    namespace = metaclass_proxy.__namespace__
    for k in getattr(v, '__provides_dependencies_for__', []):
        try:
            namespace.get_injected(k)
            return k
        except KeyError: continue
    return None

class ModelingNamespaceProxy:


    local_attributes = frozenset({
        '_metaclass',
        '_namespace',
        '_add_callback',
        '__namespace__',
        '__metaclass__',
        '__class__',
                })

    def _add_callback(self, callback):
        self._metaclass._add_callback(self._namespace, callback)

    def __init__(self, metaclass, namespace):
        self._metaclass = metaclass
        self._namespace = namespace

    @property
    def __namespace__(self): return self._namespace

    @property
    def __metaclass__(self): return self._metaclass
    
    def __getattribute__(self, attr):
        if attr in __class__.local_attributes:
            return super().__getattribute__(attr)
        try:         return self._namespace[attr]
        except KeyError:
            raise AttributeError(attr) from None
        

    def __setattr__(self, attr, value):
        if attr in __class__.local_attributes:
            return super().__setattr__(attr, value)
        self._namespace[attr] = value

    def __delattr__(self, attr):
        try: del self._namespace[attr]
        except KeyError:
            raise AttributeError(attr) from None
        
    
    

class modelmethod:

    '''Usage within the body of a :class:`ModelingBase` subclass::

        @modelmethod
def method(cls, *args, **kwargs):

    Will add ``meth`` to any class using the class where this modelmethod is added as a metaclass.  The method can be called in the class body, in which case it will receive a :class:`ModelingNamespaceProxy` rather than a completed class.  Alternatively after the class body is completed, the method can be called on the class as a classmethod, in which case it will receive the class.  It is an error to call a modelmethod on a class after it  has been instantiated.


    :param cls: The namespace proxy or class.

    Remaining parameters are passed from the call to modelmethod.

    '''

    def __init__(self, meth):
        self.method = meth

    def __repr__(self):
        return f'<modelmethod: {repr(self.method)}>'


    def __call__(self, metaclass, ns, *args, **kwargs):
        proxy = ModelingNamespaceProxy(metaclass, ns)
        return self.method(proxy, *args, **kwargs)

    def __get__(self, _class, owner):
        if _class is None: return self
        if _class._already_instantiated:
            raise AttributeError('modelmethods cannot be called after instantiation')
        return functools.partial(self.method, _class)
    
__all__ += ['modelmethod']


class ModelingBase(type):

    def _handle_decorator(target_cls, ns, k, state):
        if isinstance(state.value, decorators.ModelingDecoratorWrapper):
            return state.value.handle(target_cls, ns, k, state)

    namespace_filters: typing.List[typing.Callable] = [_handle_decorator]
    namespace_initial: typing.Mapping = {}

    @classmethod
    def __prepare__(cls, *args, **kwargs):
        classes_to_inject = combine_mro_list(cls, InjectableModelType, 'classes_to_inject')
        namespace_filters = combine_mro_list(cls, ModelingBase,
                                             'namespace_filters')
        initial = combine_mro_mapping(cls,
                                      ModelingBase,
                                      'namespace_initial')
        return ModelingNamespace(cls, filters=namespace_filters,
                                 initial=initial,
                                 classes_to_inject=classes_to_inject)

    def __new__(cls, name, bases, namespace, **kwargs):
        namespace.initial.clear()
        namespace.setitem('_already_instantiated', False)
        thread_local.current_context = None
        try:
            return super(ModelingBase, cls).__new__(cls, name, bases, namespace, **kwargs)
        except TypeError as e:
            raise TypeError(f'Error constructing ${name}: {str(e)}') from None

    def __init_subclass__(cls, *args):
        if 'namespace_filters' not in cls.__dict__:
            cls.namespace_filters = []
        if 'namespace_initial' not in cls.__dict__:
            cls.namespace_initial = {}
        for k, v in cls.__dict__.items():
            if isinstance(v, modelmethod):
                cls.namespace_initial[k] = v  # Parsed further in the namespace constructor


__all__ += ["ModelingBase"]

not_transcluded_key = InjectionKey('carthage.modeling.not_transcluded', _optional=True)
#: An injection key to store the set of keys that could have been transcluded but were not so that container propagation does not  transclude

__all__ += ['not_transcluded_key']


class InjectableModelType(ModelingBase):

    classes_to_inject: typing.Sequence[type] = (NetworkConfig, carthage.console.CarthageRunnerCommand)
    _callbacks: typing.List[typing.Callable]

    def _handle_provides(target_cls, ns, k, state):
        if hasattr(state.value, '__provides_dependencies_for__'):
            state.extra_keys.extend(state.value.__provides_dependencies_for__)
        if hasattr(state.value, '__transclusion_key__'):
            assert state.value.__transclusion_key__ in state.extra_keys, \
                f'{state.value.__transclusion_key__} must be provided by {state.value}'
            state.transclusion_key = state.value.__transclusion_key__
            

    namespace_filters = [_handle_provides]

    @classmethod
    def _add_callback(cls, ns: ModelingNamespace, cb: typing.Callable):
        ns.setdefault('_callbacks', [])
        ns['_callbacks'].append(cb)

    @classmethod
    def __prepare__(cls, name, bases, **kwargs):
        namespace = super().__prepare__(cls, name, bases, **kwargs)
        to_inject = namespace.to_inject
        transclusions = namespace.transclusions
        namespace.setitem('__transclusions__', transclusions)
        namespace.setitem('__initial_injections__', to_inject)
        for c in reversed(bases):
            if hasattr(c, '__initial_injections__'):
                to_inject.update(c.__initial_injections__)
            if hasattr(c, '__transclusions__'):
                transclusions.update(c.__transclusions__)
        return namespace

    def __new__(cls, name, bases, namespace, **kwargs):
        namespace.setdefault('_callbacks', [])

        self = super(InjectableModelType, cls).__new__(cls, name, bases, namespace, **kwargs)
        if self.__transclusions__:
            inject(_not_transcluded=not_transcluded_key)(self)

        return self

    def __init_subclass__(cls, *args, **kwargs):
        if 'classes_to_inject' not in cls.__dict__:
            cls.classes_to_inject = []
        super().__init_subclass__(*args, **kwargs)

    @modelmethod
    def add_provider(cls,  k: InjectionKey,
                     v: typing.Any = None,
                     close=True,
                     allow_multiple=False, globally_unique=False,
                     propagate=False,
                     transclusion_overrides=False,
                     force_multiple_instantiate=False):
        '''

        Similar to :meth:`Injector.add_provider` except as a model method.  *close* and *allow_multiple* work the same as on an Injector.

        :param propagate: If ``True``, then perform container propagation as this dependency moves up to ward the base injector.

        :param transclude_overrides: If ``True``,  when this :class:`InjectableModel` is instantiated, only add *k* as a provided depedency if *k* is not already in the injector hierarchy.  In effect, allow an existing provider for *k* to mask *p* at instantiation time; see :func:`transclude_overrides`.

        :param force_multiple_instantiate: Normally it is an error to call *add_provider* where the provider is itself a subclass of :class:`InjectableModel` that provides its own dependencies.  Were that to be allowed, multiple instances of the same model would be instantiated; wrapping the provider in :class:`injector_access` is almost certainly what is desired.  But if it actually is desirable to instantiate multiple instances of the same model, setting *force_multiple_instantiate* will suppress the error.

        '''

        to_inject = cls.__initial_injections__
        to_propagate = getattr(cls, '__container_propagations__', None)
        transclusions = cls.__transclusions__
        metaclass = getattr(cls, '__metaclass__', cls.__class__)
        if not isinstance(k, InjectionKey) and v is None:
            v = k
            k = default_injection_key(k)
        if globally_unique:
            warnings.warn(
                "Globally_unique is deprecated; set _globally_unique on the key",
                DeprecationWarning,
                stacklevel=2)
            k = InjectionKey(k, _globally_unique=True)
        if isinstance(v,InjectableModelType) and not force_multiple_instantiate:
            existing_key = check_already_provided(cls, v)
            if existing_key:
                raise SyntaxError(f'{v} is already provided by {existing_key}; wrap it in injector_access, or in the unlikely case that multiple instantiation is desired, set force_multiple_instantiate')
            
        if transclusion_overrides:
            transclusions[k] = {k}
        to_inject[k] = (v, dict(
            close=close,
            allow_multiple=allow_multiple,
        ))
        if propagate:
            assert issubclass(metaclass, ModelingContainer), "Only ModelingContainers accept propagation"
            to_propagate[k] = to_inject[k]

    @modelmethod
    def disable_system_dependency(cls, dependency):
        cls.__initial_injections__[dependency.default_instance_injection_key()] = (
            None, dict(close=False, allow_multiple=False))

    @modelmethod
    def self_provider(cls,  k: InjectionKey=None):
        def callback(inst):
            nonlocal k
            if k is None: k = inst.default_instance_injection_key()
            inst.injector.add_provider(k, dependency_quote(inst))
        cls._add_callback( callback)


__all__ += ['InjectableModelType']



class ModelingContainer(InjectableModelType):

    #: Returns the key under which we are registered in the parent.  Our
    # objects will be adapted by adding these constraints to register in
    # the parent as well.
    our_key: typing.Callable[[object], InjectionKey]

    def _integrate_containment(target_cls, ns, k, state):
        def propagate_provider(k_inner, v, options):
            # There is the provides_dependencies_for (or outer) keys, and there is a
            # set of inner keys from the providers being injected.  We
            # want to pick one outer key and also add the inner keys
            # (plus the constraints from the outer key) into the outer
            # injector.  We pick one outer key just for simplicity; it
            # could be made to work with more.
            globally_unique = k_inner.globally_unique
            if not globally_unique:
                if k_inner in val.__transclusions__:
                    raise TypeError(f'{k_inner} cannot be transcluded and propagated unless it is globally unique')
                outer_constraints = outer_key.constraints
                if set(outer_constraints) & set(k_inner.constraints):
                    return
                k_new = InjectionKey(k_inner.target, **outer_constraints, **k_inner.constraints)
            else:
                k_new = k_inner  # globally unique
            # Must be after we have chosen a globally unique key if there is one.
            if k_new not in ns.to_inject:
                inner_key_map[k_inner] = k_new
            if isinstance(v, decorators.injector_access):
                # Normally injector_access will not actually get a key
                # in __initial_injections__, but there are important
                # cases where that happens, like the machine entry on
                # MachineModel.  In general this situation will come
                # up if some modeling type wishes to propagate
                # something that is referenced by an InjectionKey
                # rather than a value.  One of the biggest reasons to
                # try and collapse out injector_access is that
                # injector_access cannot deal with AsyncRequired but
                # injector_xref can.
                k_inner = v.key
            # In some cases we could make the injector_xref more clear
            # by adding the outer key's constraints to
            # v.injectable_key and use that as the new injectable key,
            # preserving the inner target key.  That generally works
            # so long as the injectable_key is not masked by a key
            # already in the outer injector or by an overlapping
            # constraint.  Also, It just moves the recursion around.
            v = injector_xref(outer_key, k_inner)
            if k_new not in ns.to_inject:
                ns.to_inject[k_new] = (v, options)
                ns.to_propagate[k_new] = (v, options)

        if not isinstance(state.value, ModelingContainer):
            return
        inner_key_map = {}
        val = state.value
        outer_key = None
        if hasattr(val, '__provides_dependencies_for__'):
            for outer_key in sorted(
                    val.__provides_dependencies_for__,
                    key=lambda k: k.globally_unique,
                    reverse=True):
                if isinstance(outer_key.target, ModelingContainer) and len(outer_key.constraints) > 0:
                    break
        to_propagate = val.__container_propagations__
        if to_propagate and (outer_key is None or len(outer_key.constraints) == 0):
            warnings.warn("Cannot propagate because no outer key with constraints found", stacklevel=3)
            return
        for k, info in to_propagate.items():
            v, options = info
            propagate_provider(k, v, options)
        map_transclusions(ns.transclusions, val, inner_key_map)

    def _propagate_filter(target_cls, ns, k, state):
        if isinstance(state.value, ModelingContainer) or (state.flags & NSFlags.propagate_key):
            for k, info in ns.keys_for(name=k, state=state):
                ns.to_propagate[k] = info

    namespace_filters = [_integrate_containment, _propagate_filter]

    @classmethod
    def __prepare__(cls, *args, **kwargs):
        ns = super().__prepare__(*args, **kwargs)
        ns.to_propagate = {}
        ns.setitem('__container_propagations__', ns.to_propagate)
        return ns

    def __new__(cls, name, bases, ns, **kwargs):
        to_propagate = ns.to_propagate
        self = super().__new__(cls, name, bases, ns, **kwargs)
        to_propagate.update( sum([
            list(b.__container_propagations__.items()) for b in bases if isinstance(b, ModelingContainer)], []))
        for k in list(to_propagate.keys()):
            if k not in self.__initial_injections__:
                del self.__container_propagations[k]
        return self

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            our_key = self.our_key()
            if not isinstance(our_key, InjectionKey):
                our_key = InjectionKey(our_key)
            setattr_default(self, '__provides_dependencies_for__', [])
            self.__provides_dependencies_for__.insert(0, our_key)
        except (AttributeError, NameError):
            pass

    @modelmethod
    def include_container(cls, obj,
                          *, close=True,
                          allow_multiple=False,
                          dynamic_name=None,
                          **kwargs):
        if not hasattr(cls, '__namespace__'):
            raise TypeError('include_container can only be called within a class body')
        ns = cls.__namespace__
        def handle_function(func):
            params = set(inspect.signature(func).parameters.keys())
            open_params = params - set(kwargs.keys())
            for k in open_params:
                if k not in ns:
                    raise AttributeError(f'{k} not found in enclosing class')
                kwargs[k] = ns[k]
            return func(**kwargs)
        if not hasattr(obj, '__provides_dependencies_for__') and callable(obj):
            obj = handle_function(obj)
        if (not hasattr(obj, '__provides_dependencies_for__')) and not dynamic_name:
            raise TypeError(f'{obj} is not a modeling container class')
        state = NsEntry(obj)
        if dynamic_name:
            # This is gross, but it's not clear how to support
            # decorators ourselves
            if (not close) or allow_multiple:
                raise TypeError('If using dynamic_name, use decorators to adjust close and allow_multiple')
            ns[fixup_dynamic_name(dynamic_name)] = obj
            return
        ModelingContainer._integrate_containment(cls, ns, obj.__name__, state)
        # Since we're not able to use ns.__setitem__, do the injection key stuff ourself.
        if not close:
            state.flags &= ~NSFlags.close
        if allow_multiple:
            state.flags |= NSFlags.allow_multiple
        options = state.injection_options
        for k in obj.__provides_dependencies_for__:
            if k not in ns.to_inject:
                ns.to_inject[k] = (obj, options)
                ns.to_propagate[k] = (obj, options)


__all__ += ['ModelingContainer']


def adjust_bases_for_tasks(bases: tuple[type], namespace: dict) -> tuple[type]:
    '''
If the namespace includes any setup_tasks, then add SetupTaskMixin to the baseses.
    '''
    from ..setup_tasks import SetupTaskMixin, TaskWrapper
    if SetupTaskMixin in bases:
        return bases
    for b in bases:
        if SetupTaskMixin in b.__mro__: return bases
    for v in namespace.values():
        if isinstance(v, TaskWrapper):
            break
    else:
        return bases
    new_bases = [*bases]
    for i, c in enumerate(new_bases):
        if AsyncInjectable in c.__mro__:
            new_bases.insert(i, SetupTaskMixin)
            return tuple(new_bases)
    else:
        new_bases.extend([SetupTaskMixin,AsyncInjectable])
    return tuple(new_bases)


__all__ += ['adjust_bases_for_tasks']


def _handle_base_customization(target_cls, ns, k, state):
    val = state.value
    if not state.flags & NSFlags.instantiate_on_access:
        return
    if not isinstance(val, type):
        return
    if not issubclass(val, carthage.machine.BaseCustomization):
        return
    if state.flags & NSFlags.inject_by_class:
        key = val.default_class_injection_key()
        key = InjectionKey(carthage.machine.BaseCustomization, **key.constraints)
        ns.to_inject[key] = dependency_quote(val), state.injection_options
    state.flags &= ~(NSFlags.inject_by_class | NSFlags.instantiate_on_access)
    state.value = decorators.wrap_base_customization(val, state.new_name or k)


__all__ += ['_handle_base_customization']


def add_provider_after(cls, k: InjectionKey,
                       v: typing.Any = None,
                       close=True,
                       allow_multiple=False,
                       propagate=False,
                       transclusion_overrides=False):
    '''This is a hack two allow you to call add_provider in init_subclass.
This must happen prior to the first instantiation of a subclass, and for propagation at least currently prior to the first time the class is put in a container.
    Long term we will rework modelmethod to also add a class method usable prior to instantiation, and rework propagation so that it happens in instance initialization (or at least containing class initialization) time to give more room for adjusting.
    This will be romeved then.
    '''
    if not isinstance(k, InjectionKey) and v is None:
        v = k
        k = default_injection_key(k)
    if transclusion_overrides:
        cls.__transclusions__.add((k, k, cls))
    cls.__initial_injections__[k] = (v, dict(
        close=close,
        allow_multiple=allow_multiple,
    ))
    if propagate:
        assert isinstance(cls, ModelingContainer), "Only ModelingContainers accept propagation"
        cls.__container_propagations__[k] = cls.__initial_injections__[k]


def map_transclusions(container_transclusions, contained, inner_key_map):
    '''
    Consider the case when *contained* is added to *container*.  *contained*  provides a dependency *k*, but that dependency permits transclusion.  That means that if *contained* is instantiated in an injector  that already provides *k*, then the instantiating injector's version of *k* is used.

    It's an error for *k* not to be globally unique.  Transcluded keys must either be globally unique or not propagated, because transclusion always happens in the scope of the instantiating injector, and the only time when that naming is the same as the inner injector (for objects provided by the inner injector) is for globally unique keys.  However, when *k* is transcluded, there may be extra keys *k_1* that are also suppressed because they are provided by the same object as *k*.  These keys need to be remapped from *contained*'s transclusions to *container*'s transclusions.

    :param inner_key_map: A mapping from inner keys to outer keys.

    :param container_transclusions:  The **__transclusions__* mapping of the namespace that will become *container*.
    '''
    transclusions = container_transclusions
    inner_transclusions = contained.__transclusions__
    for transcluded_key in set(inner_transclusions) & set(inner_key_map):
        transcluded_keys = transclusions.setdefault(transcluded_key, set())
        transcluded_keys.update({inner_key_map[ik] for ik in inner_transclusions[transcluded_key]})
    
from . import decorators



# Copyright (C) 2022,  2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage.config import ConfigSchema
from .server import CarthageEntanglement, carthage_plugin
from .instrumentation import *
from . import instrumentation
from carthage import inject, Injector

class EntanglementConfig(ConfigSchema, prefix='entanglement'):

    run_server: bool = True
    ws_port:int = None
    ws_address:str = "127.0.0.1"
    port:int = 39102
    
__all__ = instrumentation.__all__
__all__ += ['CarthageEntanglement', 'carthage_plugin']



# Copyright (C) 2018, 2019, 2021, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import collections.abc
import copy
import dataclasses
import typing
from ipaddress import *
from ..dependency_injection import *


@dataclasses.dataclass()
class L3ConfigMixin:
    _attributes = frozenset({
        'dhcp',
        'dhcp_ranges',
        'address',
        'public_address',
        'network',
        'gateway',
        'secondary_addresses',
        "dns_servers", "domains",
    })

    #: Set of DNS servers that should be made available to this link/network
    dns_servers: list = None
    domains: str = None

    async def resolve(self, ainjector, interface):
        args = dict(interface=interface)
        for a in self._attributes:
            if getattr(self,a, None) is not None:
                setattr(self, a,
                        await resolve_deferred(ainjector, getattr(self, a), args))
        self.after_resolve()

    def after_resolve(self):
        if self.dhcp_ranges:
            for l, h in self.dhcp_ranges:
                if l > h:
                    raise ValueError(f'IN a dhcp range, the lower address {l} is not less than the upper address {h}')
                if self.network and l not in self.network:
                    raise ValueError(f'{l} is not in {self.network}')
                if self.network and h not in self.network:
                    raise ValueError(f'{h} is not in {self.network}')

    def _handle_dhcp_ranges(self, func):
        def wrapper(ranges):
            result = []
            for l, h in ranges:
                result.append((func(l), func(h)))
            return result
        return wrapper

    def _handle_secondary_addresses(self, func):
        def func_or_none(a):
            if a is None: return None
            if hasattr(a, 'ip_address'): a = a.ip_address
            return func(a)
        def wrapper(l):
            result = []
            for a in l:
                if not isinstance(a, (SecondaryAddress, dict)):
                    elt = SecondaryAddress(private=func(a))
                elif isinstance(a, dict):
                    elt = SecondaryAddress(
                        public=func_or_none(a.get('public')),
                        private=func_or_none(a.get('private')),
                    )
                elif isinstance(a, SecondaryAddress):
                    elt = a
                else: raise TypeError('Don\'t know how to handle secondary address'+repr(a))
                result.append(elt)
            return result
        return wrapper

    
    def _handle_pool(self, func):
        def wrapper(pool):
            assert isinstance(pool, collections.abc.Sequence)
            assert len(pool) == 2, "Format of pool is (low, high)"
            low, high = func(pool[0]), func(pool[1])
            return low, high
        return wrapper
    
    def merge(self, merge_from):
        '''
        Return a new instance of a Networkconfig where values from *merge_from* augment values not set in *self*.  Typical usage::

            merged_v4_config = link.v4_config.merge(network.v4_config)

'''
        res = copy.copy(self)
        if merge_from is None:
            return res
        for a in self._attributes:
            if getattr(res, a) is None:
                setattr(res, a, getattr(merge_from, a))
        return res

@dataclasses.dataclass()
class SecondaryAddress:
    _address = typing.Union[IPv4Address, IPv6Address, None]
    private: _address
    public: _address = None
    del _address
    

@dataclasses.dataclass()
class V4Config(L3ConfigMixin):

    network: IPv4Network = None
    dhcp: bool = None
    dhcp_ranges: list = None
    secondary_addresses: list[IPv4Address] = dataclasses.field(default_factory=lambda: [])
    address: IPv4Address = None
    gateway: IPv4Address = None
    masquerade: bool = False
    #: Takes a lower bound and a upper bound, both specified as V4 addresses.  If specified and address is None, will assign the address between the lower and upper bound.  This allows addresses to be dynamically managed at modeling time rather than by DHCP at run time.  DHCP can still be used, but at least for models whose config includes *pool*, addresses will be statically configured in the dhcp server.
    pool: tuple = dataclasses.field(default=None, repr=False)
    public_address: IPv4Address = dataclasses.field(default=None, repr=False)
    
    _attributes = L3ConfigMixin._attributes | {'masquerade', 'pool'}

    def after_resolve(self):
        # Support things like a VpcAddress being assigned to public_address
        if hasattr(self.public_address, 'ip_address'):
            self.public_address = self.public_address.ip_address
        # The following depends on iteration happening in dictionary
        # order such that network is processed before dhcp_ranges
        for k, func in dict(
                address=IPv4Address,
                network=IPv4Network,
                gateway=ipv4_gateway,
                dhcp_ranges=self._handle_dhcp_ranges(IPv4Address),
                secondary_addresses=self._handle_secondary_addresses(IPv4Address),
                pool = self._handle_pool(IPv4Address),
                public_address=ipv4_gateway,
        ).items():
            val = getattr(self, k)
            if val is not None:
                setattr(self, k, func(val))

        super().after_resolve()


def ipv4_gateway(g):
    if g is False:
        return False
    return IPv4Address(g)



# Copyright (C) 2019, 2020, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from .schema import ConfigSchema
from .types import ConfigPath, ConfigString


class BaseSchema(ConfigSchema, prefix=""):

    #: Name of layout to instantiate by default
    layout_name: str
    #: A path containing authorized keys to use in images.  If the
    # path starts with a | symbol, then the output of the given command
    # will be used.
    authorized_keys: ConfigPath = "|{hadron_operations}/hadron/inventory/config/default_keys.py"

    #: What resolv option to use when host networking is used with a container.  Likely options include bind-host, bind-stub or bind-uplink
    host_networking_resolv_conf: str = "bind-host"

    #: Ssh options that are always included
    global_ssh_options: str = '-oStrictHostKeyChecking=no'

    base_dir: ConfigPath = "~/.carthage"
    checkout_dir: ConfigPath = "{base_dir}/checkout"
    output_dir: ConfigPath = "{base_dir}/output"
    image_dir: ConfigPath = "{base_dir}"
    vm_image_dir: ConfigPath = "{base_dir}/vm"
    state_dir: ConfigPath = "{base_dir}/state"
    #: Directory for local ephemeral state like ssh_agent sockets
    local_run_dir: ConfigPath = "{state_dir}"
    vm_image_size: int = 20000000000  # : default size of VM disks in Mb
    base_container_image: str = "/usr/share/hadron-installer/hadron-container-image.tar.gz"
    base_vm_image: str = "/usr/share/hadron-installer/direct-install-efi.raw.gz"
    #: All containers and VMs are prefixed with this string in their machine names
    container_prefix: str = 'carthage-'
    min_port: int = 9000  # : Minimum port for displays and databases
    num_ports: int = 500

    #: If true, keep the ssh agent in the environment potentially
    # using production ssh keys rather than starting our own
    production_ssh_agent: bool = False
    #: Path to  a checkout of hadron_operations
    hadron_operations: ConfigPath
    delete_volumes: bool = False
    persist_local_networking: bool = False

    external_vlan_id: int = 0
    external_bridge_name: str = "brint"
    vlan_min: int = 1
    vlan_max: int = 4094


class TasksConfig(ConfigSchema, prefix="tasks"):

    #: If True, then do not actually execute tasks
    dry_run: bool = False


class DebianConfig(ConfigSchema, prefix="debian"):
    mirror: ConfigString = "http://deb.debian.org/debian"

    #: The mirror to use when running debootstrap. May be a file mirror for example to be used on the machine that will eventually be a mirror server.
    stage1_mirror: ConfigString = "{debian.mirror}"

    distribution: ConfigString = "bullseye"

    #: Whether to include security.debian.org and the updates source
    include_security: bool = True

    #: Any debootstrap option to include
    debootstrap_options: str = ""



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import argparse
import asyncio
import json
import logging
import pytest
import yaml
from. import base_injector, ConfigLayout
from .dependency_injection import AsyncInjector


@pytest.fixture(scope='session')
def loop():
    ''':returns: Asyncio event loop
'''
    return asyncio.get_event_loop()


@pytest.fixture(scope='session')
def test_parameters(pytestconfig):
    try:
        return pytestconfig.carthage_test_parameters
    except AttributeError:
        pytest.skip("Test parameters not specified")


def pytest_collection_modifyitems(items):
    # This hook modifies items wrapped by @async_test to add fixtures used by the wrapped function
    # See the comment in that code for details

    for i in items:
        if isinstance(i, pytest.Function):
            if hasattr(i.function, '__signature__'):
                del i.keywords._markers['place_as']
                del i.keywords._markers['usefixtures']
                del i.keywords._markers['__signature__']
                i._fixtureinfo.argnames = tuple(i.function.__signature__.parameters.keys())


@pytest.fixture()
def ainjector():
    ainjector = base_injector.claim()(AsyncInjector)
    yield ainjector
    ainjector.close()


def pytest_addoption(parser):
    group = parser.getgroup("Carthage", "Carthage Continuous Integration Options")
    group.addoption('--carthage-config',
                    type=argparse.FileType('rt'),
                    help="Specify yaml carthage config; this configuration file describes where to put VMs and where to find hadron-operations.  It is not the test configuration for individual tests.  This option is typically used on the controller and not alongside --carthage-json on the system under test.",
                    metavar="file")
    group.addoption('--carthage-json',
                    metavar="file",
                    type=argparse.FileType('wt'),
                    help="Write json results to this file")
    group.addoption('--test-parameters', '--test-params',
                    metavar='file',
                    type=argparse.FileType('rt'),
                    help="YAML test parameters.  This option is typically used on the system under test alongside --carthage-json and not typically used with --carthage-config."
                    )
    group.addoption('--carthage-commands-verbose',
                    action='store_true',
                    help='Enable verbose logging of all carthage commands run')


def pytest_configure(config):
    global json_out
    global json_log
    json_log = config.getoption('carthage_json')
    json_out = []
    if not config.getoption('carthage_commands_verbose'):
        logging.getLogger('sh').setLevel(logging.ERROR)
        logging.getLogger('carthage.sh').propagate = False
        carthage_config = config.getoption('carthage_config')
    if carthage_config:
        config_layout = base_injector(ConfigLayout)
        try:
            config_layout.load_yaml(carthage_config)
        finally:
            carthage_config.close()
    test_params_yaml = config.getoption('test_parameters')
    if test_params_yaml:
        config.carthage_test_parameters = yaml.load(test_params_yaml)
        test_params_yaml.close()


def pytest_runtest_logreport(report):
    if json_log is None:
        return
    d = {}
    for k in ('nodeid', 'location', 'keywords', 'outcome', 'longrepr', 'when', 'sections', 'duration'):
        d[k] = getattr(report, k)
    d['longrepr'] = report.longreprtext
    json_out.append(d)


def pytest_sessionfinish():
    global json_out, json_log
    if json_log is None:
        return
    json_log.write(json.dumps(json_out))
    json_out = []
    json_log.close()
    json_log = None



# Copyright (C) 2018, 2019, 2020, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import argparse
import asyncio
import contextlib
import fcntl
import functools
import logging
import os
import pathlib
import re
import typing
import weakref
import importlib.resources
import mako.lookup


async def possibly_async(r):
    '''If r is a coroutine, await it.  Otherwise return it.  Used like the
    following:

        return await possibly_async(self.check_volume())

    check_volume can now optionally be declared async
    '''
    if asyncio.iscoroutine(r):
        return await r
    else:
        return r


class memoproperty:
    "A property that only supports getting and that stores the result the first time on the instance to avoid recomputation"

    def __init__(self, fun):
        functools.update_wrapper(self, fun)
        self.fun = fun
        self.name = fun.__name__

    def __get__(self, instance, owner):
        if instance is None:
            return self
        # Because we don't define set or del, we should not be called
        # if name is already set on instance.  So if we set name we
        # will be bypassed in the future
        res = self.fun(instance)
        setattr(instance, self.name, res)
        return res


class WhenNeededMeta(type):

    def __repr__(self):
        if self.resolved_obj:
            return f'<when_needed resolved={repr(self.resolved_obj)}>'

        wraps, args, kwargs, injector = self.repr_info
        if isinstance(wraps, type):
            wraps_repr = wraps.__name__
        else:
            wraps_repr = repr(wraps)
        s = "when_needed({}".format(wraps_repr)
        for a in args:
            s += ", {}".format(a)
        for k, v in kwargs.items():
            s += ", {}={}".format(k, v)
        s += ")"
        if injector is not None:
            s += ", injector ={}".format(repr(injector))
            s += ")"
        return s


def when_needed(wraps, *args, injector=None,
                addl_keys=[],
                **kwargs):
    '''Return an AsyncInjectable class that when resolved will instantiate
    'wraps' giving it *args and **kwargs.  This is only done once; if this
    class is instantiated more than once the same shared object is
    returned.  By default, the injector used when we are instantiated will
    be used, but if the injector keyword is passed in then that injector
    will always be used for the instantiation.  Differences between this
    and passing a type into add_provider are:

    * If the return from this function is passed into add_provider for
      different injection keys, then these objects will be the same
      when instantiated.  Without when_needed, the keys associated
      with each call to add_provider would instantiate different
      objects of the same type.

    * This may be used to instantiate objects that require
      non-injected configuration to their constructors.

    '''
    from .dependency_injection import inject, AsyncInjectable, AsyncInjector, InjectionKey, Injectable, _call_close, Injector
    # We do not copy the wrapped function's dependencies out.  We will
    # submit the wrapped object to an injector as part of resolving it
    # and we may need to control which injector is used for the
    # dependencies.
    # We override the injector dependency to avoid autokwargs behavior because we do not want our injector claimed

    @inject(injector=Injector)
    @functools.wraps(wraps,
                     assigned=functools.WRAPPER_ASSIGNMENTS,
                     updated=tuple())
    class WhenNeeded(AsyncInjectable, metaclass=WhenNeededMeta):

        resolved_obj = None
        resolving = None

        def __init__(self, *inside_args, **inside_kwargs):
            nonlocal args
            if args and inside_args:
                raise RuntimeError(
                    "It does not make sense to specify args both in the call to when_needed and when it is resolved.")
            if inside_args:
                args = inside_args
            if injector is not None:
                # override injector
                inside_kwargs['injector'] = injector
            # We want to end up  with a copied injector but not one that is claimed so
            # the resulting object can claim it.
            self.injector = inside_kwargs.pop('injector')
            self.injector = self.injector(type(self.injector))
# super will set up ainjector
            super().__init__()
            self.ainjector_set.add(self.ainjector)
            self.inside_kwargs = inside_kwargs

        @classmethod
        def supplementary_injection_keys(self, k):
            if isinstance(wraps, type) and issubclass(wraps, Injectable):
                yield from wraps.supplementary_injection_keys(k)
            yield from addl_keys

        @classmethod
        def close(self, canceled_futures=None):
            if self.resolved_obj:
                _call_close(self.resolved_obj, canceled_futures)
            for ainjector in self.ainjector_set:
                ainjector.close(canceled_futures=canceled_futures)
            self.ainjector_set.clear()
            if hasattr(self, 'resolving') and self.resolving:
                self.resolving.cancel()
                if canceled_futures:
                    canceled_futures.append(self.resolving)
                self.resolving = None

        # class level property to track all the injectors in use.
        ainjector_set = weakref.WeakSet()
        repr_info = (wraps, args, kwargs, injector)

        async def async_resolve(self):
            nonlocal kwargs
            if self.resolved_obj:
                return self.resolved_obj
            if self.resolving:
                return await asyncio.shield(self.resolving)
            loop = self.ainjector.get_instance(asyncio.AbstractEventLoop)
            self.__class__.resolving = loop.create_future()
            del loop
            kws = kwargs.copy()
            kws.update(self.inside_kwargs)
            try:
                res = await self.ainjector(wraps, *args, **kws)
                self.__class__.resolved_obj = res
                self.resolving.set_result(res)
                del self.__class__.resolving
                # We will never need them again so release the references
                kwargs = None
                del self.__class__.repr_info
                del self.ainjector
                del self.inside_kwargs
                return res
            except Exception as e:
                self.resolving.set_exception(e)
                self.__class__.resolving = None  # try again next time
                raise

    addl_keys = list(map(
        lambda k: k if isinstance(k, InjectionKey) else InjectionKey(k), addl_keys))

    return WhenNeeded


def permute_identifier(id, maxlen):
    "Add to or replace the last character of the identifier; use as generator and stop consuming when a unique one is found"
    yield id
    if len(id) < maxlen:
        for i in range(10):
            yield id + chr(97 + i)
    else:
        id = id[:-1]
        for i in range(10):
            yield id + chr(97 + i)
    raise ValueError("No unique combination found")


def add_carthage_arguments(parser):
    parser.add_argument('--config',
                        metavar="file",
                        default=[],
                        type=argparse.FileType('rt'),
                        action='append')
    parser.add_argument('--command-verbose',
                        help="Verbose command logging",
                        action=argparse.BooleanOptionalAction)
    parser.add_argument('--tasks-verbose',
                        help="Verbose logging for tasks",
                        action=argparse.BooleanOptionalAction)
    parser.add_argument('--plugin',
                        dest='plugins',
                        default=[],
                        action='append',
                        help='Load a plugin into Carthage',
                        metavar='plugin')
    return parser


def carthage_main_argparser(*args, **kwargs):
    parser = argparse.ArgumentParser(*args, **kwargs)
    add_carthage_arguments(parser)
    return parser


def carthage_main_setup(parser=None, unknown_ok=False, ignore_import_errors=False):
    from . import base_injector, ConfigLayout
    from .plugins import load_plugin
    if parser is None:
        parser = carthage_main_argparser()
    if unknown_ok:
        result = parser.parse_known_args()
        args = result[0]
    else:
        args = parser.parse_args()
        result = args
    if len(args.config) > 0:
        config = base_injector(ConfigLayout)
        for f in args.config:
            config.load_yaml(f, ignore_import_errors=ignore_import_errors)
    for p in args.plugins:
        base_injector(load_plugin, p, ignore_import_errors=ignore_import_errors)
    root_logger = logging.getLogger()
    console_handler = logging.StreamHandler()
    root_logger.addHandler(console_handler)
    root_logger.setLevel('INFO')
    container_logger = logging.getLogger('carthage.container')
    container_logger.addHandler(logging.FileHandler('container.log'))
    container_logger.setLevel(10)

    def container_debug_filter(record):
        if record.name == 'carthage.container' and record.levelno == 10:
            return 0
        return 1
    console_handler.addFilter(container_debug_filter)
    if not args.command_verbose:
        logging.getLogger('sh').setLevel(logging.ERROR)
        logging.getLogger('carthage.sh').propagate = False
        logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)
        logging.getLogger('urllib3.connectionpool').propagate = False
    if args.tasks_verbose:
        logging.getLogger('carthage.setup_tasks').setLevel(10)

    return result


def carthage_main_run(func, *args, **kwargs):
    loop = asyncio.get_event_loop()
    from . import base_injector, AsyncInjector, shutdown_injector
    from .config import inject_config
    inject_config(base_injector)
    ainjector = base_injector(AsyncInjector)
    try:
        return loop.run_until_complete(ainjector(func, *args, **kwargs))
    finally:
        loop.run_until_complete(shutdown_injector(base_injector))


shell_safe_re = re.compile(r'^[a-zA-Z0-9_./ ]*$')


def validate_shell_safe(s):
    if shell_safe_re.search(s):
        return True
    return False


@contextlib.contextmanager
def TemporaryMountPoint(**kwargs):
    '''
    Create a tempory directory to be used as a mount point.  The directory name is the value of the context.
    If the directory is empty, it is deleted when the context is exited.

    The advantage of this function over :class:`tempfile.TemporaryDirectory` for mount points is that if somehow the unmount fails, the mounted filesystem's contents are not deleted.
'''
    dir = tempfile.mkdtemp(**kwargs)
    try:
        yield dir
    finally:
        os.rmdir(dir)


def import_resources_files(package):
    "stub for importlib.resources.files"
    try:
        return importlib.resources.files(package)
    except AttributeError:
        if isinstance(package, str):
            return pathlib.Path(importlib.import_module(package).__path__[0])
        else:
            return pathlib.Path(package.__path__[0])


mako_lookup = mako.lookup.TemplateLookup([import_resources_files(__package__) / "resources/templates"],
                                         strict_undefined=True)


def is_optional_type(t):
    # support both python 3.7 and python 3.9
    # As of python 3.9 there is not a way to do this without using internals
    try:
        if t.__class__ == typing._UnionGenericAlias:
            t = typing.get_args(t)
            if type(None) in t:
                return True
    except AttributeError:
        if t.__class__ is typing._GenericAlias:
            if type(None) in t.__args__:
                return True
    return False


def get_type_args(t):
    try:
        return typing.get_type_args(t)
    except AttributeError:
        return t.__args__


@contextlib.asynccontextmanager
async def file_locked(fd: typing.Union[int, str], mode=fcntl.LOCK_EX, unlock=False):
    loop = asyncio.get_event_loop()
    close = False
    if isinstance(fd, str) or hasattr(fd, "__fspath__"):
        fd = os.open(fd, os.O_CREAT | os.O_CLOEXEC | os.O_RDWR, 0o664)
        close = True

    def lock(m):
        fcntl.lockf(fd, m)
    await loop.run_in_executor(None, lock, mode)
    try:
        yield
    finally:
        if unlock:
            await loop.run_in_executor(None, lock, fcntl.LOCK_UN)
        if close:
            os.close(fd)


class NotPresentType:
    '''A singleton value indicating that an dependency should not be injected into a function if if the dependency is not provided.  Used as the value for the *_optional* parameter to :func:`~carthage.dependency_injection.inject`
    Usage::

        @inject(router=InjectionKey(Router, _optional=NotPresent))
        def func(**kwargs):
            # If the injector does not provide Router, kwargs will not contain 'router'
            #If _optional had been True, then router would be None when not provided.

    '''

    def __new__(cls):
        raise TypeError('NotPresent is a singleton')

    def __repr__(self):
        return 'NotPresent'


NotPresent = object.__new__(NotPresentType)


def relative_path(p):
    '''Returns a :class:`pathlib.Path` that is guaranteed to be relative.  Intended to be used to make sure that joining the return value to a filesystem root does not escape the filesystem.
    '''
    p = pathlib.Path(p)
    if p.is_absolute():
        return p.relative_to('/')
    return p


__all__ = ['when_needed', 'possibly_async', 'permute_identifier', 'memoproperty',
           'add_carthage_arguments', 'carthage_main_argparser',
           'carthage_main_setup', 'carthage_main_run',
           'validate_shell_safe',
           'is_optional_type',
           'TemporaryMountPoint',
           'import_resources_files',
           'mako_lookup',
           'file_locked',
           'relative_path',
           ]



# Copyright (C)  2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from .base import *
from .base import InjectorClosed, _call_close, is_obj_ready, instantiate_to_ready
from .introspection import *

__all__ = [
    'AsyncInjectable', 'AsyncInjector', 'AsyncRequired',
    'DependencyProvider',
    'ExistingProvider', 'Injectable', 'InjectionFailed',
    'InjectionKey', 'Injector', 'InstantiationContext', 'aspect_for',
    'NotPresent',
    'current_instantiation', 'dependency_quote', 'inject',
    'resolve_deferred',
    'inject_autokwargs', 'injector_xref',
    'partial_with_dependencies', 'shutdown_injector',
    'injection_failed_unlogged', 'instantiation_not_ready',
    'instantiate_to_ready'
]



# Copyright (C) 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio

from entanglement import SyncServer

from .instrumentation import carthage_registry, CarthageDestination

from carthage.dependency_injection import *
from carthage import ConfigLayout


@inject_autokwargs(injector=Injector,
                   loop=asyncio.AbstractEventLoop,
                   )
class CarthageEntanglement(Injectable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.loop.call_later(0.2, self.start_server)

    def start_server(self):
        self.config_layout = self.injector(ConfigLayout)
        config = self.config_layout.entanglement
        if config.run_server or config.ws_port:
            self.server = SyncServer(cert=None, port=config.port,
                                     registries=[carthage_registry],
                                     loop=self.loop)

        if config.ws_port:
            import tornado.web
            import tornado.httpserver
            from entanglement.websocket import SyncWsHandler
            self.web_app = tornado.web.Application([(r'/entanglement_ws', SyncWsHandler)])
            self.http_server = tornado.httpserver.HTTPServer(self.web_app)
            self.http_server.listen(int(config.ws_port), address=str(config.ws_address))
            self.web_app.sync_manager = self.server
            self.web_app.find_sync_destination = self.websocket_destination

    def websocket_destination(self, request, *args, **kwargs):
        dest = CarthageDestination()
        return dest
    

@inject(injector=Injector)
def carthage_plugin(injector):
    injector.add_provider(CarthageEntanglement)
    injector.get_instance(CarthageEntanglement)
    carthage_registry.instrument_injector(injector)
    



# Copyright (C) 2019, 2020, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import sys
from ipaddress import IPv6Address, IPv4Address, IPv4Network, IPv6Network
from pathlib import Path
from ..dependency_injection import inject, Injectable, InjectionKey, Injector, partial_with_dependencies, InjectorClosed, InjectionFailed, injection_failed_unlogged


def config_key(k):
    return InjectionKey("config/" + k)


class ConfigResolutionFailed(ValueError):

    def __init__(self, k, val):
        self.config_key = k
        self.config_val = val
        super().__init__(f'Resolution of {k} with value `{val}` failed')


class ConfigSchemaMeta(type):

    def __new__(mcls, name, bases, namespace, *, prefix, **kwargs):
        cls = type.__new__(mcls, name, bases, namespace, **kwargs)
        if prefix != "" and not prefix.endswith('.'):
            prefix += '.'
        cls._schema = mcls._find_schema(prefix)
        try:
            annotations = cls._eval_annotations(cls.__annotations__)
        except AttributeError:
            annotations = {}
            # We do not yet parse docs
        schema = cls._schema
        for k in annotations:
            if k.startswith('_'):
                continue
            default = getattr(cls, k, None)
            if isinstance(default, dict):
                raise TypeError(
                    "A dict as a default for a configuration value is probably too confusing; discuss whether this is really what you want.")
            if k in schema:
                raise TypeError(f'{prefix}{k} is already defined')
            schema[k] = cls.Item(prefix + k, type_=annotations[k],
                                 default=default)
        return cls

    def subsections(cls, prefix):
        subsections = set()
        # prefix ends in a trailing period
        for k in cls._schemas:
            if k is prefix:
                continue
            if k.startswith(prefix):
                next = k[len(prefix):]
                next, sep, tail = next.partition('.')
                if next not in subsections:
                    subsections.add(next)
                    yield next

    @classmethod
    def _find_schema(cls, prefix):
        schemas = cls._schemas  # start at the root
        prefix = prefix.rstrip('.')
        try:
            return schemas[prefix]
        except KeyError:
            schemas[prefix] = {}
            return schemas[prefix]

    _schemas = {}

    def _eval_annotations(cls, annotations):
        annotations = annotations.copy()
        locls = cls.__dict__
        globls = sys.modules[cls.__module__].__dict__
        for k, v in annotations.items():
            if isinstance(v, str):
                v = eval(v, globls, locls)
                if not isinstance(v, type):
                    raise TypeError(f'`{v}` must evaluate to a type')
                annotations[k] = v
        return annotations

    def __repr__(cls):
        d = {k: v.default for k, v in cls._schema.items()}
        return f'<{cls.__name__}: {d}>'


class ConfigSchema(metaclass=ConfigSchemaMeta, prefix=""):
    '''
    Class representing the valid options in a Carthage Configuration schema

    Typical usage::

        class  VmwareConfig(ConfigSchema, prefix = "vmware"):

            #: Which vMware datacenter should be used?
            datacenter: str

        class VmwareHardware(ConfigSchema, prefix = "vmware.hardware"):

            memory: int = 128 #: memory In megabytes
            paravirt: bool = False
    '''

    class Item:

        '''
        Represents an item in a configuration schema
'''

        __slots__ = ('name', 'type', 'default', 'key')

        def __init__(self, name, type_, default):
            assert isinstance(type_, type), f'{name} config key must be declared with a type not {type_}'
            if type_ is bool:
                from .types import ConfigBool
                type_ = ConfigBool
            elif type_ is str:
                from .types import ConfigString
                type_ = ConfigString
            self.name = name
            self.type = type_
            self.default = default
            self.key = config_key(name)

        def __repr__(self):
            return f'ConfigSchema.Item("{self.name}", {self.type.__name__}, {repr(self.default)})'

        def resolve(self, injector):
            "Return the value of this item resolved against the given injector"
            try:
                res = injector.get_instance(self.key)
                return res
            except InjectorClosed:
                # We may miss substitutions, but this is the best we can do if the injector is closed
                return self.default
                return res
            except (KeyError):
                if self.default is None:
                    return None
                try:
                    with injection_failed_unlogged():
                        res = injector(self.type, self.default)
                    return res
                except Exception as e:
                    raise ConfigResolutionFailed(self.name, self.default) from None


@inject(injector=Injector)
class ConfigAccessor:

    def __init__(self, injector, prefix):
        if not (prefix == "" or prefix.endswith('.')):
            prefix = prefix + '.'
        self._injector = injector
        self._prefix = prefix
        try:
            self._schema = ConfigSchema._schemas[prefix.rstrip('.')]
        except KeyError:
            raise KeyError(f'{prefix} is not a valid configuration prefix') from None

    def __getattr__(self, k):
        if k.startswith('_'):
            raise AttributeError
        try:
            return self._schema[k].resolve(self._injector)
        except KeyError:
            if self._prefix + k in ConfigSchema._schemas:
                try:
                    return self._injector.get_instance(config_key(self._prefix + k))
                except KeyError:
                    return self._injector(ConfigAccessor, prefix=self._prefix + k)
            raise AttributeError(f'{self._prefix}{k} is not a valid configuration key') from None

    def __setattr__(self, k, v):
        if k.startswith('_'):
            return super().__setattr__(k, v)
        if not hasattr(self, k):
            raise AttributeError("{} is not a configuration key".format(self._prefix + k))
        if isinstance(v, dict):
            raise ValueError("You cannot set a configuration key to a dictionary")
        self._injector.replace_provider(config_key(self._prefix + k), v)

    def _dictify(self, include_defaults=False):
        from .types import ConfigBool
        d = {}
        for k, schema_item in self._schema.items():
            def_v = schema_item.default
            try:
                v = getattr(self, k)
            except (ConfigResolutionFailed, InjectionFailed):
                v = "<resolution failed>"
            if include_defaults or (v != def_v):
                for t in (float, int, str):
                    # bool is an int but we want to handle that below
                    if isinstance(v, t) and not isinstance(v, (bool, ConfigBool)):
                        v = t(v)  # Remove any ConfigValue class
                        break
                else:
                    if isinstance(v, ConfigBool):
                        v = bool(v)
                    elif isinstance(v, (tuple, list)):
                        v = list(v)
                    elif v is None or isinstance(v, bool):
                        # just fine as is
                        pass
                    elif isinstance(v, (IPv4Address, IPv4Network, IPv6Address, IPv6Network, Path)):
                        v = str(v)
                    else:
                        raise TypeError(f'{v} is a {type(v)} which will not work so well saved to YAML')

                d[k] = v

        for k in ConfigSchema.subsections(self._prefix):
            d_subsection = getattr(self, k)._dictify(include_defaults=include_defaults)
            if d_subsection or include_defaults:
                d[k] = d_subsection
        return d

    def __getstate__(self):
        return self._dictify(True)

    def __repr__(self):
        return f'<{self.__class__.__name__} overrides: {self._dictify()}>'



# Copyright (C) 2019, 2020, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import contextlib
import dataclasses
import json
import os
import os.path
import tempfile
import typing
import yaml
import importlib.resources
from abc import ABC, abstractmethod
from pathlib import Path

from .dependency_injection import *
from . import sh, Machine, machine
from .container import Container
from .config import ConfigLayout
from .ssh import SshKey
from .utils import validate_shell_safe
from types import SimpleNamespace
from .network import access_ssh_origin
from . import setup_tasks
from .plugins import CarthagePlugin
import logging

logger = logging.getLogger("carthage")
__all__ = []


class AnsibleFailure(RuntimeError):

    def __init__(self, msg, ansible_result=None):
        super().__init__(msg)
        self.ansible_result = ansible_result

    def __str__(self):
        s = super().__str__()
        if self.ansible_result:
            s += ": " + repr(self.ansible_result)
        return s


__all__ += ['AnsibleFailure']

ansible_origin = InjectionKey("ansible/origin")
ansible_log = InjectionKey("ansible/log")

__all__ += ['ansible_origin', "ansible_log"]


@inject(injector=Injector)
class AnsibleConfig(Injectable):
    '''
    Capture carthage plugins providing ansible resources etc.
    '''

    def __init__(self, injector):
        # If this class is ever modified to store the injector, then it should be
        # passed into the superclass so it can be claimed.
        super().__init__()
        roles = []
        filters = []
        for k, pl in injector.filter_instantiate(CarthagePlugin, ['name']):
            roles_path = pl.resource_dir / "ansible/roles"
            if roles_path.exists():
                roles += [str(roles_path)]
            filter_path = pl.resource_dir / "ansible/filter_plugins"
            if filter_path.exists():
                filters.append(str(filter_path))
        self.roles = roles
        self.filter_plugins = filters


__all__ += ['AnsibleConfig']


@dataclasses.dataclass
class NetworkNamespaceOrigin:
    '''An *origin* to be passed into :func:`.run_playbook` that indicates
    that we wish to run *Ansible* using the network namespace to which
    a :class:`.Container` belongs.  The container could be used
    directly, but in that case *run_playbook* will ssh to the
    container, effectively using both the network and filesystem
    namespace.  In contrast, with this class, the filesystem namespace
    in which *Carthage* runs will be used.
    '''
    namespace: Container


__all__ += ["NetworkNamespaceOrigin"]


class AnsibleInventory(AsyncInjectable):

    '''A source of ansible inventory.  This class will generate an
        ansible inventory file and optionally transfer it to a target
        :class:`~carthage.machine.Machine`.

        :param destination: A file path or :class:`carthage.ssh.RsyncPath` where the resulting inventory yaml will be placed.


    '''

    def __init__(self, destination, **kwargs):
        self.destination = destination
        super().__init__(**kwargs)

    async def async_ready(self):
        result = await self.generate_inventory()
        await self.write_inventory(self.destination, result)
        await super().async_ready()

    async def collect_machines(self):
        self.machines = await self.ainjector.filter_instantiate_async(Machine, ['host'], ready=False)

    async def collect_groups(self):
        plugins = await self.ainjector.filter_instantiate_async(AnsibleGroupPlugin, ['name'], ready=True)
        plugins = sorted(plugins, key=lambda x: getattr(x[0], 'priority', 100), reverse=True)
        self.group_plugins = [p[1] for p in plugins]
        group_info: dict[str, dict] = {}
        for k, p in plugins:
            try:
                result = await self.ainjector(p.group_info)
            except BaseException:
                logger.exception(f"Error getting group variables from {p.name} plugin:")
                raise
            for group, result_info in result.items():
                group_info.setdefault(group, {})
                for info_type, info_type_dict in result_info.items():
                    group_info[group].setdefault(info_type, {})
                    # info_type will be vars, hosts or children
                    for k, v in info_type_dict.items():
                        group_info[group][info_type][k] = v
        return group_info

    async def collect_hosts(self, result_dict: dict[str, dict]):
        plugin_filtered = await self.ainjector.filter_instantiate_async(AnsibleHostPlugin, ['name'], ready=True)
        plugin_filtered = sorted(plugin_filtered, key=lambda x: getattr(x[0], 'priority', 100), reverse=True)
        plugins = [p[1] for p in plugin_filtered]
        all = result_dict.setdefault('all', {})
        hosts_dict = all.setdefault('hosts', {})
        for ignore_key, m in self.machines:
            try:
                machine_name = m.ansible_inventory_name
            except AttributeError:
                machine_name = m.name
            var_dict: dict[str, dict] = {}
            for p in plugins:
                try:
                    var_dict.update(await self.ainjector(p.host_vars, m))
                except BaseException:
                    logger.exception(f"Error getting variables for {machine_name} from {p.name} plugin:")
                    raise
            if 'ansible_host' not in var_dict:
                try:
                    var_dict['ansible_host'] = m.ip_address
                except BaseException:
                    pass
            hosts_dict[machine_name] = var_dict
            groups = []
            for p in self.group_plugins:
                try:
                    groups += await self.ainjector(p.groups_for, m)
                except BaseException:
                    logger.exception(f"Error determining groups for {machine_name} from group plugin {p.name}")
                    raise
            for g in groups:
                result_dict.setdefault(g, {})
                result_dict[g].setdefault('hosts', {})
                result_dict[g]['hosts'].setdefault(machine_name, {})

    async def generate_inventory(self):
        await self.collect_machines()
        result = await self.collect_groups()
        await self.collect_hosts(result)
        self.inventory = result
        return result

    async def write_inventory(self, destination, inventory: dict[str, dict]):
        from . import ssh
        with contextlib.ExitStack() as stack:
            if not isinstance(destination, ssh.RsyncPath):
                local_path = Path(destination)
                os.makedirs(local_path.parent, exist_ok=True)
            else:
                dir = stack.enter_context(
                    TemporaryDirectory(
                        dir=self.config_layout.state_dir,
                        prefix="ansible-inventory"))
                local_path = os.path.join(dir, "hosts.yml")
            with open(local_path, "wt") as f:
                f.write(yaml.dump(inventory, default_flow_style=False))
                if isinstance(destination, ssh.RsyncPath):
                    key = await self.ainjector.get_instance_async(ssh.SshKey)
                    await key.rsync(local_path, destination)


class AnsibleGroupPlugin(Injectable, ABC):

    @abstractmethod
    async def group_info(self) -> dict[str: dict[str: typing.Any]]:
        '''
        Returns an ansible inventory dictionary.  An example might look like the following in yaml::

            group_name:
                vars:
                    var_1: value_1
                children:
                    group_2:
                hosts:
                    host.com:

        '''
        raise NotImplementedError

    @abstractmethod
    async def groups_for(self, m: Machine):
        raise NotImplementedError

    @classmethod
    def default_class_injection_key(self):
        return InjectionKey(AnsibleGroupPlugin, name=self.name)


class AnsibleHostPlugin(AsyncInjectable, ABC):

    @abstractmethod
    async def host_vars(self, m: Machine):
        raise NotImplementedError

    @classmethod
    def default_class_injection_key(self):
        return InjectionKey(AnsibleHostPlugin, name=self.name)


__all__ += ['AnsibleInventory', 'AnsibleGroupPlugin', 'AnsibleHostPlugin']


@inject(origin=InjectionKey(ansible_origin, optional=True),
        inventory=AnsibleInventory,
        log=InjectionKey(ansible_log, optional=True),
        ainjector=AsyncInjector,
        ansible_config=AnsibleConfig,
        config=ConfigLayout,
        )
async def run_playbook(hosts,
                       playbook: str,
                       inventory: str,
                       log: str = None,
                       *,
                       extra_args=[],
                       raise_on_failure: bool = True,
                       ansible_config,
                       ainjector,
                       config,
                       origin=None,
                       ):

    '''
    Run an ansible playbook against a set of hosts.
    A new ansible configuration is created for each run of ansible.
    if *log* is ``None`` then ansible output is parsed and an :class:`AnsibleResult` is returned.  Otherwise ``True`` is returned on a successful ansible run.

    :param playbook: Path to the playbook local to *origin*.  The current directory when ansible is run will be the directory containing the playbook.

    :param hosts: A list of hosts to run the play against.  Hosts can be strings, or :class:`Machine` objects.  For machine objects, *ansible_inventory_name* will be tried before :meth:`Machine.name`.  If a machine is not running and has :meth:`ansible_not_running_context`, then that asyncronous context manager will be entered.  The return from that context manager will be used as inventory variables for the machine.  For example :class:`~carthage.container.Container` uses this to arrange to run ansible plays without booting a container.

:param log: A local path where a log of output and errors should be created or ``None`` to parse the ansible result programatically after ansible completes.  It is not possible to produce human readable output and a result that can be parsed at the same time without writing a custom ansible callback plugin.

:param origin: Controls where the playbook is run:

    None
        Run locally

    A :class:`~carthage.machine.Machine`
        If the origin is a :class:`Container` that is not running, run ansible-playbook directly in the namespace of the container.  Otherwise ssh into the machine and run ansible-playbook.

    NetworkNamespaceOrigin
        Use the local filesystem but the network namespace of the referenced container.

    '''
    injector = ainjector.injector

    def to_inner(s):
        s = str(s)
        return str(config_inner) + s[len(str(config_dir)):]
    if not isinstance(hosts, list):
        hosts = [hosts]
    async with contextlib.AsyncExitStack() as stack:
        target_hosts = []
        inventory_overrides = {}
        for h in hosts:
            if isinstance(h, str):
                target_hosts.append(h)
            else:
                if hasattr(h, 'ansible_inventory_name'):
                    target_name = h.ansible_inventory_name
                else:
                    target_name = h.name
                if not h.running and hasattr(h, 'ansible_not_running_context'):
                    inventory_overrides[target_name] = await stack.enter_async_context(
                        h.ansible_not_running_context())
                elif hasattr(h, 'ansible_inventory_overrides'):
                    inventory_overrides[target_name] = h.ansible_inventory_overrides
                target_hosts.append(target_name)

        list(map(lambda h: validate_shell_safe(h), target_hosts))
        target_hosts = ":".join(target_hosts)
        playbook = str(playbook)
        playbook_dir = os.path.dirname(playbook)
        validate_shell_safe(playbook)
        if isinstance(inventory, AnsibleInventory):
            inventory = inventory.destination
        if playbook_dir:
            ansible_command = f'cd "{playbook_dir}"&& '
        else:
            ansible_command = ''
        if origin and not isinstance(origin, NetworkNamespaceOrigin):
            config_dir = await stack.enter_async_context(origin.filesystem_access())
            config_dir = config_dir / "ansible"
            os.makedirs(config_dir, exist_ok=True)
            config_inner = "/ansible"
        else:
            config_dir = config.state_dir
            config_inner = config.state_dir
        if isinstance(origin, Container) and not origin.running:
            # This may not be strictly true, but we definitely
            # cannot get access to the json
            log = "container.log"
        with await ainjector(
                write_config, config_dir,
                [inventory],
                ansible_config=ansible_config, log=log,
                inventory_overrides=inventory_overrides,
                origin=origin) as config_file:

            ansible_command += f'ANSIBLE_CONFIG={to_inner(config_file)} ansible-playbook -l"{target_hosts}" {" ".join(extra_args)} {os.path.basename(playbook)}'
            log_args: dict = {}
            if log:
                log_args['_out'] = log
                log_args['_err_to_out'] = True

            if origin and isinstance(origin, NetworkNamespaceOrigin):
                await stack.enter_async_context(origin.namespace.machine_running(ssh_online=True))
                cmd = injector(access_ssh_origin, ssh_origin=origin.namespace)(
                    "sh",
                    '-c', ansible_command,
                    **log_args,
                    _bg=True,
                    _bg_exc=False)
            elif origin:
                vrf = origin.injector.get_instance(InjectionKey(machine.ssh_origin_vrf, optional=True))
                if vrf:
                    ansible_command = ansible_command.replace("ansible-playbook",
                                                              f'ip vrf exec {vrf} ansible-playbook')
                if isinstance(origin, Container) and not origin.running:
                    cmd = origin.container_command("sh", "-c", ansible_command)
                else:
                    cmd = origin.ssh("-A", ansible_command,
                                     _bg=True,
                                     _bg_exc=False,
                                     **log_args)
            else:
                cmd = sh.sh(
                    '-c', ansible_command,
                    **log_args,
                    _bg=True,
                    _bg_exc=False)
            try:
                ansible_exc = None
                await cmd
            except (sh.ErrorReturnCode_2, sh.ErrorReturnCode_4) as e:
                # Remember the exception, preferring it to other
                # exceptions if we get a parse error on the output
                ansible_exc = e
            except Exception as e:
                if log:
                    log_str = f'; logs in {log}'
                else:
                    log_str = ""
                raise AnsibleFailure(f'Failed Running {playbook} on {target_hosts}{log_str}') from e
        if log and ansible_exc:
            raise AnsibleFailure(f'Failed running {playbook} on {target_hosts}; logs in {log}') from ansible_exc
        elif log:
            return True
        try:
            json_out = json.loads(cmd.stdout)
            result = AnsibleResult(json_out)
        except Exception:
            if ansible_exc:
                raise ansible_exc from None
            raise
        if (not result.success) and raise_on_failure:
            raise AnsibleFailure(f'Failed running {playbook} on{target_hosts}', result)
        return result

__all__ += ['run_playbook']


@inject(ainjector=AsyncInjector,
        inventory=InjectionKey(AnsibleInventory, _optional=True),
        origin=InjectionKey(ansible_origin, _optional=True),
        log=InjectionKey(ansible_log, _optional=True),
        )
async def run_play(hosts, play,
                   *, raise_on_failure=True,
                   gather_facts=False,
                   vars=None, inventory=None,
                   log=None,
                   origin=None,
                   ainjector):
    '''
    Run a single Ansible play, specified as a python dictionary.
    The typical usage of this function is for cases where code wants to use an Ansible module to access some resource, especially when the results need to be programatically examined.  As an example, this can be used to examine the output of Ansible modules that gather facts.

    **If you are considering loading a YAML file, parsing it, and calling this function, you are almost certainly better served by :func:`run_playbook`.**
'''
    async with contextlib.AsyncExitStack() as stack:
        if origin and not isinstance(origin, NetworkNamespaceOrigin):
            root_path = await stack.enter_async_context(origin.filesystem_access())
            ansible_dir = stack.enter_context(tempfile.TemporaryDirectory(dir=root_path, prefix="ansible-"))
        else:
            ansible_dir = stack.enter_context(tempfile.TemporaryDirectory())
            root_path = "/"
        ansible_dir = Path(ansible_dir)
        with open(os.path.join(ansible_dir,
                               "playbook.yml"), "wt") as f:
            if isinstance(play, dict):
                play = [play]
            pb = [{
                'hosts': ":".join([h.name for h in hosts]),
                'remote_user': 'root',
                'gather_facts': gather_facts,
                'tasks': play}]
            if callable(vars):
                vars = await ainjector(vars)
            if vars:
                pb[0]['vars'] = vars
            f.write(yaml.dump(pb, default_flow_style=False))
        if inventory is None:
            with open(os.path.join(ansible_dir,
                                   "inventory.txt"), "wt") as f:
                f.write("[hosts]\n")
                for h in hosts:
                    if not h.running and hasattr(h, 'ansible_not_running_context'):
                        continue
                    try:
                        f.write(h.ansible_inventory_line() + "\n")
                    except AttributeError:
                        try:
                            f.write(f'{h.name} ansible_ip={h.ip_address}\n')
                        except (NotImplementedError, AttributeError):
                            pass
            inventory = f'/{ansible_dir.relative_to(root_path)}/inventory.txt'
        return await ainjector(run_playbook,
                               hosts,
                               f'/{ansible_dir.relative_to(root_path)}/playbook.yml',
                               inventory=inventory,
                               raise_on_failure=raise_on_failure,
                               log=log,
                               origin=dependency_quote(origin))

__all__ += ['run_play']


@inject(ssh_key=InjectionKey(SshKey, _optional=True),
        config=ConfigLayout)
@contextlib.contextmanager
def write_config(config_dir, inventory,
                 *, ssh_key, log,
                 config, origin,
                 inventory_overrides,
                 ansible_config):
    private_key = ""
    if ssh_key:
        private_key = f"private_key_file = {ssh_key.key_path}"
    if not log:
        stdout_str = "stdout_callback = json"
    else:
        stdout_str = ""
    with tempfile.TemporaryDirectory(
            dir=config_dir,
            prefix="ansible") as dir, \
            open(dir + "/ansible.cfg", "wt") as f:
        f.write("[defaults]\n")
        f.write(f'''\
{stdout_str}
retry_files_enabled = false
roles_path={":".join(ansible_config.roles)}
filter_plugins={":".join(ansible_config.filter_plugins)}
{private_key}

''')
        if inventory_overrides or origin is None or isinstance(origin, NetworkNamespaceOrigin):
            os.makedirs(dir + "/inventory/group_vars")
            with open(dir + "/inventory/hosts.yml", "wt") as hosts_file:
                # This may need to be more generalized as Carthage
                # core gains the ability to construct nontrivial
                # Ansible inventory.  Another concern is that the way
                # Ansible uses hosts, groups and variables is
                # conceptually similar to how Carthage uses injectors.
                # It's likely that we want a way to reference more
                # injection keys as dependencies for a play than just
                # config variables, and to access attributes of those
                # dependencies.  If those mechanisms become available,
                # we may want to revisit this.  This hosts file is
                # sometimes empty if there are no inventory_overrides,
                # but is still needed to make a valid inventory
                # source.
                hosts_file.write(yaml.dump(
                    dict(all=dict(hosts=inventory_overrides)),
                    default_flow_style=False))

            with open(dir + "/inventory/group_vars/all.yml", "wt") as config_yaml:
                config_yaml.write(yaml.dump(
                    {'config': config.__getstate__()},
                    default_flow_style=False))
            inventory = [dir + "/inventory/hosts.yml"] + inventory
        f.write(f'inventory = {",".join(inventory)}')

        # ssh section
        f.write('''
[ssh_connection]
pipelining=True
''')
        if origin is None or isinstance(origin, NetworkNamespaceOrigin):
            f.write(f'''\
ssh_args = -o ControlMaster=auto -o ControlPersist=60s -oUserKnownHostsFile={config.state_dir}/ssh_known_hosts {config.global_ssh_options}
''')
        else:
            f.write(f'''\
ssh_args = -o ControlMaster=auto -o ControlPersist=60s -oStrictHostKeyChecking=no
''')

        f.flush()
        yield dir + "/ansible.cfg"


class LocalhostMachine:
    name = "localhost"
    ip_address = "127.0.0.1"
    running = True

    def ansible_inventory_line(self):
        return "localhost ansible_connection=local"


localhost_machine = LocalhostMachine()

__all__ += ['localhost_machine']


class AnsibleResult:

    def __init__(self, res):
        self.json = res
        self.failures = 0
        self.unreachable = 0
        self.changed = 0
        self.ok = 0
        for v in res['stats'].values():
            self.ok += v['ok']
            self.failures += v['failures']
            self.changed += v['changed']
            self.unreachable += v['unreachable']
        self.parse_plays(res)
        self.host_stats = res['stats']

    def parse_plays(self, res):
        self.tasks = {}
        for p in res['plays']:
            for t in p['tasks']:
                t.update(t['task'])
                del t['task']
                t['duration'] = SimpleNamespace(**t['duration'])
                if t['name'] in self.tasks:
                    raise ValueError(f"{t['name']} duplicated")
                if len(t['hosts']) == 1:
                    for v in t['hosts'].values():
                        t.update(v)
                self.tasks[t['name']] = SimpleNamespace(**t)

    @property
    def success(self):
        return (self.unreachable == 0) and (self.failures == 0)

    def __repr__(self):
        res = f"<AnsibleResult: \
failures: {self.failures}; unreachable: {self.unreachable}; ok: {self.ok}; changed: {self.changed};\
plays:[{[k for k in self.tasks.keys()]}]"

        if self.failures:
            for t in self.tasks.values():
                try:
                    if t.failed:
                        res += f"\n\tFatal {t.name}: {t.msg}"
                except AttributeError:
                    pass
        res += ">"
        return res


__all__ += ['AnsibleResult']


def _handle_host_origin(host, origin):
    if not isinstance(host, Machine) and hasattr(host, 'host'):
        host = host.host
    if not origin:
        return host, {}
    if isinstance(host, Container) and not host.running:
        return localhost_machine, dict(origin=dependency_quote(host))
    return host, dict(origin=dependency_quote(host))


class ansible_playbook_task(setup_tasks.TaskWrapper):

    extra_attributes = frozenset({'dir', 'playbook'})

    def __init__(self, playbook, origin=False, **kwargs):
        async def func(inst):
            host, extra_args = _handle_host_origin(inst, origin)
            return await inst.ainjector(run_playbook, host, self.dir.joinpath(self.playbook), **extra_args)
        super().__init__(
            func=func,
            description=f'Run {playbook} playbook',
            **kwargs)
        self.playbook = playbook
        self.dir = None

    def __set_name__(self, owner, name):
        import sys
        module = sys.modules[owner.__module__]
        try:
            self.dir = importlib.resources.files(module.__package__)
        except (AttributeError, ValueError):
            self.dir = Path(module.__file__).parent


__all__ += ['ansible_playbook_task']


def ansible_role_task(roles, vars=None,
                      before=None, order=None, origin=False):
    '''
    A :func:`setup_task` to apply one or more ansible roles to a machine.

    :param roles: A single role (as a string) or list of roles to include.  Roles can also be a list of dictionaries containing argumens to *import_role*.

    :param vars: An optional dictionary of ansible variable assignments.
    :param origin: If True, use host as origin from which to run ansible.

    '''
    @setup_tasks.setup_task(f'Apply {roles} ansible roles',
                            before=before, order=order)
    @inject(ainjector=AsyncInjector)
    async def apply_roles(self, ainjector):
        host, extra_args = _handle_host_origin(self, origin)
        play = []
        for r in roles:
            if isinstance(r, dict):
                r_dict = r
            else:
                r_dict = dict(name=r)
            play.append(dict(
                import_role=r_dict))

        return await ainjector(
            run_play,
            hosts=[host],
            play=play,
            vars=vars,
            **extra_args
        )
    if isinstance(roles, (str, dict)):
        roles = [roles]
    return apply_roles


__all__ += ['ansible_role_task']

@inject(model=machine.AbstractMachineModel)
def ansible_log_for_model(model):
    '''
    used like::
        add_provider(ansible_log, ansible_log_for_model, allow_multiple=True)

    Sets up a per-model ansible log in *stamp_path*/ansible.log.
    '''
    return f'{model.stamp_path}/ansible.log'

__all__ += ['ansible_log_for_model']



# Copyright (C) 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from ..config import ConfigSchema, config_key


class CredentialsSchema(ConfigSchema, prefix="vmware"):
    hostname: str
    username: str = "{vault:secret/password/{vmware.hostname}:username}"
    password: str = "{vault:secret/password/{vmware.hostname}:password}"
    validate_certs: bool = False


#: Injection key to get Vmware credential configuration.  Only assume that username, hostname, password and verify_certs are set with this key as a dependency.
vmware_credentials = config_key('vmware')

__all__ = ['vmware_credentials']



# Copyright (C) 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import dataclasses
import os
import tempfile
import yaml
from pathlib import Path
from .dependency_injection import *
from .network import NetworkConfig, NetworkLink
from .machine import AbstractMachineModel
from . import sh, ConfigLayout
import carthage.ssh

__all__ = []


@dataclasses.dataclass
class CloudInitConfig:

    meta_data: dict = dataclasses.field(default_factory=lambda: {})
    user_data: dict = dataclasses.field(default_factory=lambda: {
        'disable_root': False})

    network_configuration: dict = dataclasses.field(default_factory=lambda: {})

    __all__ += ['CloudInitConfig']


@inject_autokwargs(model=AbstractMachineModel)
class CloudInitPlugin(Injectable):

    async def apply(self, config: CloudInitConfig):
        "Apply changes to the given :class:`CloudInitConfig`.  This is typically overridden so the plugin actually does something."
        pass

    @classmethod
    def default_class_injection_key(cls):
        return InjectionKey(CloudInitPlugin, name=cls.name)

    @property
    def name(self):
        raise NotImplementedError


__all__ += ['CloudInitPlugin']


@inject(model=AbstractMachineModel,
        ainjector=AsyncInjector)
async def generate_cloud_init_cloud_config(*, ainjector, model):
    '''
    Generate a :class:`CloudInitConfig` from the model.  The operation is based on the *cloud_init* property in the model:

    * True: Generate cidata based on calling all the :class:`CloudInitPlugins <CloudInitPlugin>` registered with the model's injector.

    * False: return None

    * A :class:`CloudInitConfig` object: use that object without calling plugins.

    '''

    cloud_init = getattr(model, 'cloud_init', False)
    if not cloud_init:
        return
    if cloud_init is True:
        config = CloudInitConfig()
        plugin_data = await ainjector.filter_instantiate_async(CloudInitPlugin, ['name'], ready=True)
        for plugin_key, plugin in plugin_data:
            # we apply in order so that plugins can look at previous
            # results.  We sacrifice parallelism to get this.
            await plugin.apply(config)
        if not plugin_data:
            return
        return config
    else:
        return model.cloud_init

__all__ += ['generate_cloud_init_cloud_config']


@inject(model=AbstractMachineModel, ainjector=AsyncInjector,
        config_layout=ConfigLayout)
async def generate_cloud_init_cidata(
        *, model, ainjector, config_layout):
    '''

    Generates a cidata ISO imagebased on the provided model.  Operation is based on *cloud_init* in the model:

    * True: Generate cidata based on calling all the :class:`CloudInitPlugins <CloudInitPlugin>` registered with the model's injector.

    * False: return None

    * A :class:`CloudInitConfig` object: use that object without calling plugins.

    :return: Path to an ISO under the *stamp_path* of *model*

    '''
    config = await generate_cloud_init_cloud_config(ainjector=ainjector, model=model)
    if not config:
        return
    with tempfile.TemporaryDirectory(dir=config_layout.state_dir) as tmp_dir:
        tmp = Path(tmp_dir)
        with tmp.joinpath("user-data").open("wt") as f:
            f.write("#cloud-config\n")
            f.write(yaml.dump(config.user_data))
        with tmp.joinpath("meta-data").open("wt") as f:
            f.write(yaml.dump(config.meta_data))
        if config.network_configuration:
            config.network_configuration.update(version=2)
            with tmp.joinpath("network-config").open("wt") as f:
                f.write(yaml.dump(config.network_configuration))
        output = Path(model.stamp_path) / "cloud_init.iso"
        output_tmp = Path(model.stamp_path) / "cloud_init.iso.tmp"
        await sh.genisoimage(
            "-J", "--rational-rock",
            "-o", output_tmp,
            "-V", "cidata",
            str(tmp),
            _bg=True, _bg_exc=False)
        os.rename(output_tmp, output)

    return output

__all__ += ['generate_cloud_init_cidata']


class NetworkPlugin(CloudInitPlugin):

    name = "network"

    async def apply(self, config: CloudInitConfig):
        ethernets = config.network_configuration.setdefault('ethernets', {})
        for l in self.model.network_links.values():
            if l.local_type:
                continue
            v4_config = l.merged_v4_config
            if l.member_of_links:
                member = l.member_of_links[0]
                if (member.member_links[0] == l) and (member.local_type in ('bridge', 'bond')):
                    v4_config = member.merged_v4_config
            if not (v4_config.dhcp or v4_config.address):
                continue
            eth_dict = dict()
            if l.mac:
                eth_dict['match'] = dict(macaddress=l.mac)
                eth_dict['set-name'] = l.interface
            if v4_config.dhcp:
                eth_dict['dhcp4'] = True
            else:
                eth_dict['dhcp4'] = False
                if v4_config.address:
                    eth_dict.setdefault('addresses', [])
                    eth_addresses = eth_dict['addresses']
                    eth_addresses.append(str(v4_config.address) + '/' + str(v4_config.network.prefixlen))

            ethernets[l.interface] = eth_dict


@inject_autokwargs(authorized_keys=carthage.ssh.AuthorizedKeysFile)
class AuthorizedKeysPlugin(CloudInitPlugin):

    name = "ssh_keys"

    async def apply(self, config: CloudInitConfig):
        authorized_keys_file = Path(self.authorized_keys.path)
        authorized_keys = list(filter(
            lambda k: k and (not k[0] == '#'),
            authorized_keys_file.read_text().split("\n")))
        config.meta_data['public_ssh_keys'] = authorized_keys


class HostnamePlugin(CloudInitPlugin):

    name = "hostname"

    async def apply(self, config: CloudInitConfig):
        config.user_data['hostname'] = self.model.name


@inject_autokwargs(authorized_keys=carthage.ssh.AuthorizedKeysFile)
class WriteAuthorizedKeysPlugin(CloudInitPlugin):

    '''
            This plugin uses the write_files module to write out root's authorized keys file.  It is not enabled by default because metadata is better when it is available.  But for example with EC2, this may be desirable to write out the full Carthage authorized_keys file without  dealing with EC2 keypairs.
    '''

    name = "write_authorized_keys"

    async def apply(self, config):
        write_files = config.user_data.setdefault('write_files', [])
        with open(self.authorized_keys.path, 'rt') as f:
            content = f.read()
        write_files.append(dict(
            path="/root/.ssh/authorized_keys",
            content=content,
            permissions='0644',
            owner='root:root'))


@inject(injector=Injector)
def enable_cloud_init_plugins(injector):
    injector.add_provider(AuthorizedKeysPlugin, allow_multiple=True)
    injector.add_provider(NetworkPlugin, allow_multiple=True)
    injector.add_provider(HostnamePlugin, allow_multiple=True)


__all__ += ['enable_cloud_init_plugins']



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

from pyVim.connect import Connect, SmartConnect, Disconnect
from ssl import create_default_context

import logging


@inject(config=ConfigLayout)
class VmwareConnection(Injectable):

    def __init__(self, config):
        self.config = config.vmware
        ssl_context = create_default_context()
        if self.config.validate_certs is False:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = 0
        self.connection = None
        logging.debug(f'connecting to {self.config.hostname} as {self.config.username}')
        self.connection = SmartConnect(host=self.config.hostname,
                                       user=self.config.username,
                                       pwd=self.config.password,
                                       sslContext=ssl_context)
        self.content = self.connection.content
        logging.debug(f'connected to {self.config.hostname}')

    def close(self):
        if self.connection:
            Disconnect(self.connection)
            self.connection = None

    def __del__(self):
        self.close()
